"""
CV.py
====================================
Class for performing train test splits and other cross validation for BPt
"""
import sklearn.model_selection as MS
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator


def get_bpt_cv(cv, dataset):
    '''Helper function used to return BPtCV object from passed
    cv params object and dataset.'''

    # Process the passed cv strategy according to dataset
    cv_strategy = dataset._proc_cv_strategy(cv.cv_strategy)

    # Check for is split corresponds to a leave-out-group
    if isinstance(cv.splits, str):

        if not dataset._is_category(cv.splits, check_scopes=False):
            raise RuntimeError('Passed splits col must have scope category.')

        dataset._check_roles()
        if dataset.roles[cv.splits] != 'non input':
            raise RuntimeError('Passed splits col must be role == non input.')

        # @TODO input validation on choice of leave out variable
        splits_vals = dataset[cv.splits]

    else:
        splits_vals = None

    return BPtCV(splits=cv.splits, n_repeats=cv.n_repeats,
                 cv_strategy=cv_strategy, splits_vals=splits_vals,
                 random_state=cv.random_state, only_fold=cv.only_fold)


class BPtCV(BaseEstimator):

    def __init__(self, splits, n_repeats,
                 cv_strategy, splits_vals,
                 random_state='context', only_fold=None):

        self.splits = splits
        self.n_repeats = n_repeats
        self.cv_strategy = cv_strategy
        self.splits_vals = splits_vals
        self.random_state = random_state
        self.only_fold = only_fold

    def get_cv(self, fit_index, random_state, return_index):

        if self.random_state != 'context':
            random_state = self.random_state

        return self.cv_strategy.get_cv(fit_index,
                                       splits=self.splits,
                                       n_repeats=self.n_repeats,
                                       splits_vals=self.splits_vals,
                                       random_state=random_state,
                                       return_index=return_index,
                                       only_fold=self.only_fold)

    def get_split(self, fit_index, random_state):
        '''Used when there is only one split, returns as index'''

        # Make sure splits is a float
        if not isinstance(self.splits, float):
            raise RuntimeError('splits must be a float to use get_split!')

        if self.random_state != 'context':
            random_state = self.random_state

        return self.cv_strategy.train_test_split(fit_index,
                                                 test_size=self.splits,
                                                 random_state=random_state,
                                                 return_index=True)

    def get_n_repeats(self):

        if self.only_fold is None:
            return self.n_repeats
        return 1


def inds_from_names(original_subjects, subject_splits):

    subject_inds = [[[original_subjects.get_loc(name) for name in s]
                     for s in split] for split in subject_splits]
    return subject_inds


class CVStrategy(BaseEstimator):
    '''Class that holds a CV_Behavior'''

    def __init__(self, groups=None, stratify=None, train_only=None):
        '''
        CV class init, for defining split behavior. If groups is None
        and stratify is None, random splits are used.

        Parameters
        ----------
        groups : pandas Series or None, optional
            `groups` should be passed a pandas Series,
            by default passed from ML.strat,
            where the Index are subject id's and the values are
            unique ids - where like ids are to be preserved within
            the same fold.
            E.g., GroupKFold

            (default = None)

        stratify : pandas Series or None, optional
            `groups` should be passed a pandas Series,
            by default passed from ML.strat,
            where the Index are subject id's and the values are
            unique ids - where like ids are to be as evenly distributed
            as possible between folds.
            E.g., StratifiedKFold

            (default = None)

        train_only : array-like, optional
            An array of subjects that should be forced into the
            training set at every fold, or train test split.

            (default = None)

        '''

        self.groups = groups
        self.stratify = stratify
        self.train_only = train_only

        if self.train_only is not None:
            if len(self.train_only) == 0:
                self.train_only = None

    def __repr__(self):
        return 'CV()'

    def __str__(self):
        return self.__repr__()

    def repeated_train_test_split(self, subjects, n_repeats, test_size=.2,
                                  random_state=None, return_index=False):

        subject_splits = []
        for n in range(n_repeats):

            # If a specific random state is passed
            # Make sure each repeat is provided a different random_state
            if random_state is not None:
                random_state += 1

            subject_splits.append(self.train_test_split(
                subjects,
                test_size=test_size,
                random_state=random_state,
                return_index=return_index))

        return subject_splits

    def train_test_split(self, subjects, test_size=.2,
                         random_state=None, return_index=False):
        '''Define a train test split on input subjects, with a given target
        test size.

        Parameters
        ----------
        subjects : array-like
            `subjects` should be a pandas index or numpy array of subjects.
            They should correspond to any subject indexed groups or stratify.

        test_size : float, int or None, optional
            If float, should be between 0.0 and 1.0 and represent
            the proportion of the dataset to be included in the test split.
            If int, represents the absolute number (or target number) to
            include in the testing group.
            (default = .2)

        random_state : int or None, optional
            Optionally can provide a random state, in
            order to be able to recreate exact splits.
            (default=None)

        Returns
        ----------
        array-like
            The training subjects as computed by the split

        array-like
            The testing subjects as computed by the split
        '''

        original_subjects, subjects, train_only = self.get_train_only(subjects)

        if self.groups is not None:
            splitter = MS.GroupShuffleSplit(n_splits=1, test_size=test_size,
                                            random_state=random_state)
            [*inds] = splitter.split(subjects,
                                     groups=self.groups.loc[subjects])

        elif self.stratify is not None:
            splitter = MS.StratifiedShuffleSplit(n_splits=1,
                                                 test_size=test_size,
                                                 random_state=random_state)
            [*inds] = splitter.split(subjects,
                                     y=self.stratify.loc[subjects])

        else:
            splitter = MS.ShuffleSplit(n_splits=1, test_size=test_size,
                                       random_state=random_state)
            [*inds] = splitter.split(subjects)

        inds = inds[0]

        train_subjects, test_subjects = subjects[inds[0]], subjects[inds[1]]
        train_subjects = np.concatenate([train_subjects, train_only])

        if return_index:
            return ([original_subjects.get_loc(name)
                     for name in train_subjects],
                    [original_subjects.get_loc(name)
                     for name in test_subjects])

        return train_subjects, test_subjects

    def repeated_k_fold(self, subjects, n_repeats,
                        n_splits, random_state=None,
                        return_index=False):
        '''Perform a repeated k-fold with class defined split behavior.
        This function simply calls a repeated version of self.k_fold.

        Parameters
        ----------
        subjects : array-like
            `subjects` should be a pandas index or numpy array of subjects.
            They should correspond to any subject indexed groups or stratify.

        n_repeats : int
            The number of times to repeat the k-fold split.

        n_splits : int
            The number of splits to compute within each repeated split.
            Or rather the k in k-fold.

        random_state : int or None, optional
            Optionally can provide a random state, in
            order to be able to recreate exact splits.
            (default = None)

        return_index : bool, optional
            If true, then instead of returning subject ID's, this function
            will return the numerical index of the subject.
            (default = False)

        Returns
        ----------
        list of tuples of array-like
            This function returns a list,
            where each element in this outer list is another fold.
            E.g., if n_repeats = 2, and n_splits = 2, then this list will
            have a length of 4. Within each list,
            there is a tuple with two elements.
            The first index of the tuple contains the array-like
            list of either subjects or numerical index
            (This depends on the value of return index)
            of the training subjects in that fold.
            The second contains the testing subjects.
        '''

        subject_splits = []
        for n in range(n_repeats):

            # If a specific random state is passed
            # Make sure each repeat is provided a different random_state
            if random_state is not None:
                random_state += 1

            subject_splits += self.k_fold(subjects, n_splits, random_state,
                                          return_index)

        return subject_splits

    def k_fold(self, subjects, n_splits, random_state=None,
               return_index=False):
        '''Perform a k-fold with class defined split behavior.

        Parameters
        ----------
        subjects : array-like
            `subjects` should be a pandas index or numpy array of subjects.
            They should correspond to any subject indexed groups or stratify.

        n_splits : int
            The number of splits to compute within each fold.
            Or rather the k in k-fold.

        random_state : int or None, optional
            Optionally can provide a random state, in
            order to be able to recreate exact splits.
            (default = None)

        return_index : bool, optional
            If true, then instead of returning subject ID's, this function
            will return the numerical index of the subject.
            (default = False)

        Returns
        ----------
        list of tuples of array-like
            This function returns a list,
            where each element in this outer list is another fold.
            E.g., if n_splits = 2, then this list will have a length of 2.
            Within each list, there is a tuple with two elements.
            The first index of the tuple contains the array-like
            list of either subjects or numerical index
            (This depends on the value of return index)
            of the training subjects in that fold.
            The second contains the testing subjects.
        '''

        original_subjects, subjects, train_only = self.get_train_only(subjects)

        # Special implementation for group K fold,
        # just do KFold on unique groups, and recover subjects
        # by group. This assume groups are roughly equal size.
        if self.groups is not None:

            groups = self.groups.loc[subjects]
            unique_groups = np.unique(groups)

            splitter = MS.KFold(n_splits=n_splits, shuffle=True,
                                random_state=random_state)

            [*inds] = splitter.split(unique_groups)

            subject_splits = [(
                np.concatenate([groups.index[groups.isin(unique_groups[i[0]])],
                                train_only]),
                groups.index[groups.isin(unique_groups[i[1]])]) for i in inds]

        else:

            # Stratify behavior just uses stratified k-fold
            if self.stratify is not None:

                splitter = MS.StratifiedKFold(n_splits=n_splits,
                                              shuffle=True,
                                              random_state=random_state)
                [*inds] = splitter.split(subjects,
                                         y=self.stratify.loc[subjects])

            # If no groups or stratify, just use random k-fold
            else:

                splitter = MS.KFold(n_splits=n_splits, shuffle=True,
                                    random_state=random_state)
                [*inds] = splitter.split(subjects)

            # Conv inds within subjects to subject names + train only subjects
            subject_splits = [(np.concatenate([subjects[i[0]], train_only]),
                               subjects[i[1]]) for i in inds]

        if return_index:
            return inds_from_names(original_subjects, subject_splits)

        return subject_splits

    def get_train_only(self, subjects, ignore_by_group=False):

        original_subjects = subjects.copy()

        # Make sure original_subjects is pandas index
        if not isinstance(original_subjects, pd.Index):
            original_subjects = pd.Index(original_subjects)

        if self.train_only is not None:

            # Calculate train only
            train_only = subjects.intersection(self.train_only)

            # If groups, then train only also includes any subjects
            # with the same group ID. Though don't apply for e.g.
            # leave_one_group_out, since could lead to funky behavior.
            if self.groups is not None and not ignore_by_group:

                groups = self.groups.loc[subjects]

                train_only_unique_groups =\
                    np.unique(self.groups.loc[train_only])

                train_only =\
                    groups[groups.isin(train_only_unique_groups)].index

            # Set subjects as the difference between subjects and train only
            subjects = subjects.difference(train_only)

        else:
            train_only = pd.Index(data=[])

        return original_subjects, subjects, train_only

    def repeated_leave_one_group_out(self, subjects, n_repeats, groups_series,
                                     return_index=False):

        subject_splits = []
        for _ in range(n_repeats):
            subject_splits += self.leave_one_group_out(subjects, groups_series,
                                                       return_index)

        return subject_splits

    def leave_one_group_out(self, subjects, groups_series, return_index=False):

        original_subjects, subjects, train_only =\
            self.get_train_only(subjects, ignore_by_group=True)

        logo = MS.LeaveOneGroupOut()
        [*inds] = logo.split(subjects, groups=groups_series.loc[subjects])

        subject_splits = [(np.concatenate([subjects[i[0]], train_only]),
                           subjects[i[1]]) for i in inds]

        if return_index:
            return inds_from_names(original_subjects, subject_splits)

        return subject_splits

    def get_num_groups(self, subjects, groups_series):
        '''Func to get number of leave one out groups'''

        _, subjects, _ =\
            self.get_train_only(subjects, ignore_by_group=True)
        groups = groups_series.loc[subjects]

        return len(np.unique(groups))

    def get_cv(self, fit_index, splits, n_repeats,
               splits_vals=None, random_state=None,
               return_index=False, only_fold=None):
        '''Always return as list of tuples'''

        if return_index == 'both':
            no_index = self.get_cv(fit_index, splits, n_repeats,
                                   splits_vals=splits_vals,
                                   random_state=random_state,
                                   return_index=False)

            index = self.get_cv(fit_index, splits, n_repeats,
                                splits_vals=splits_vals,
                                random_state=random_state,
                                return_index=True)

            return no_index, index

        # If split_vals passed, then by group
        if splits_vals is not None:

            cv_inds =\
                self.repeated_leave_one_group_out(fit_index,
                                                  n_repeats=n_repeats,
                                                  groups_series=splits_vals,
                                                  return_index=return_index)

        # K-fold is splits is an int
        elif isinstance(splits, int):

            cv_inds =\
                self.repeated_k_fold(fit_index, n_repeats,
                                     n_splits=splits,
                                     random_state=random_state,
                                     return_index=return_index)

        # Otherwise, as train test splits
        else:

            cv_inds =\
                self.repeated_train_test_split(fit_index, n_repeats,
                                               test_size=splits,
                                               random_state=random_state,
                                               return_index=return_index)

        return proc_only_fold(cv_inds, only_fold)


def proc_only_fold(cv_inds, only_fold):

    if isinstance(only_fold, int):
        only_fold = [only_fold]

    if only_fold is None:
        return cv_inds

    elif isinstance(only_fold, list):

        cv_subset_inds = []

        for of in only_fold:

            try:
                cv_subset_inds.append(cv_inds[of])
            except IndexError:
                raise IndexError(f'Padded only_fold={of} is out ' +
                                 'of range for the generated CV folds.')

        return cv_subset_inds

    raise RuntimeError(f'only_fold={only_fold} is invalid.')



from sklearn.utils.metaestimators import _BaseComposition
from sklearn.utils.metaestimators import if_delegate_has_method
from ..default.params.Params import Dict, Choice
from .base import _get_est_fit_params, _get_est_trans_params


class Selector(_BaseComposition):

    _needs_mapping = True
    _needs_fit_index = True
    _needs_transform_index = True

    def __init__(self, estimators, to_use=0):
        self.estimators = estimators
        self.to_use = to_use
        self.example_estimator_ = self.estimators[0][1]

    def get_params(self, deep=True):
        return self._get_params('estimators', deep=deep)

    def set_params(self, **kwargs):

        # Pass params as dict with key select
        select = kwargs['select']

        # Get to use from select
        self.to_use = select['to_use']

        # Set rest of select params
        self._set_params('estimators', **select)
        return self

    @if_delegate_has_method(delegate='example_estimator_')
    def fit(self, *args, **kwargs):

        # Select correct estimator
        self.estimator_ = self.estimators[self.to_use][1]

        # Set correct fit params based on chosen estimator
        fit_params = _get_est_fit_params(self.estimator_, other_params=kwargs)

        # Fit
        self.estimator_.fit(*args, **fit_params)
        self.is_fitted_ = True

        return self

    @if_delegate_has_method(delegate='example_estimator_')
    def fit_transform(self, *args, **kwargs):

        self.estimator_ = self.estimators[self.to_use][1]
        return self.estimator_.fit_transform(*args, **kwargs)

    @if_delegate_has_method(delegate='example_estimator_')
    def transform(self, *args, **kwargs):

        if 'transform_index' in kwargs:
            transform_index = kwargs.pop('transform_index')
        else:
            transform_index = None

        tranform_params = _get_est_trans_params(self.estimator_,
                                                transform_index)

        return self.estimator_.transform(*args, **tranform_params)

    @if_delegate_has_method(delegate='example_estimator_')
    def fit_resample(self, *args, **kwargs):
        self.estimator_ = self.estimators[self.to_use][1]
        return self.estimator_.fit_resample(*args, **kwargs)

    @if_delegate_has_method(delegate='example_estimator_')
    def fit_predict(self, *args, **kwargs):
        self.estimator_ = self.estimators[self.to_use][1]
        return self.estimator_.fit_predict(*args, **kwargs)

    @if_delegate_has_method(delegate='estimator_')
    def predict(self, *args, **kwargs):
        return self.estimator_.predict(*args, **kwargs)

    @if_delegate_has_method(delegate='estimator_')
    def predict_proba(self, *args, **kwargs):
        return self.estimator_.predict_proba(*args, **kwargs)

    @if_delegate_has_method(delegate='estimator_')
    def decision_function(self, *args, **kwargs):
        return self.estimator_.decision_function(*args, **kwargs)

    @if_delegate_has_method(delegate='estimator_')
    def predict_log_proba(self, *args, **kwargs):
        return self.estimator_.predict_log_proba(*args, **kwargs)

    @if_delegate_has_method(delegate='estimator_')
    def score(self, *args, **kwargs):
        return self.estimator_.score(*args, **kwargs)

    @if_delegate_has_method(delegate='estimator_')
    def inverse_transform(self, *args, **kwargs):
        return self.estimator_.inverse_transform(*args, **kwargs)

    @if_delegate_has_method(delegate='estimator_')
    def transform_df(self, *args, **kwargs):
        return self.estimator_.transform_df(*args, **kwargs)

    @if_delegate_has_method(delegate='estimator_')
    def _proc_new_names(self, *args, **kwargs):
        return self.estimator_._proc_new_names(*args, **kwargs)

    @property
    def _estimator_type(self):
        '''This should remain static across all passed estimators'''
        return self.example_estimator_._estimator_type

    @property
    def feature_importances_(self):
        if hasattr(self.estimator_, 'feature_importances_'):
            return getattr(self.estimator_, 'feature_importances_')
        return None

    @property
    def coef_(self):
        if hasattr(self.estimator_, 'coef_'):
            return getattr(self.estimator_, 'coef_')
        return None

    @property
    def classes_(self):
        if hasattr(self.estimator_, 'classes_'):
            return getattr(self.estimator_, 'classes_')
        return None


def selector_wrapper(objs, params, name):

    selector = (name, Selector(objs))

    p_dicts = []
    for i in range(len(objs)):
        obj_name = objs[i][0]
        rel_params =\
            {p: params[p] for p in params if p.split('__')[0] == obj_name}
        rel_params['to_use'] = i

        p_dict = Dict(**rel_params)
        p_dicts.append(p_dict)

    select = Choice(p_dicts, deterministic=True)
    select_params = {name + '__select': select}

    return selector, select_params



from ..data_file_type import DataFileDtype
from ..data_file import DataFile
from ..Dataset import Dataset
import numpy as np
import tempfile
import os
import pandas as pd


def test_basic():

    temp_dr = tempfile.gettempdir()

    data_files = []
    for i in range(3):
        file_loc = os.path.join(temp_dr, 'fake' + str(i) + '.npy')
        data = np.arange(1, 10)
        np.save(file_loc, data)
        data_files.append(DataFile(loc=file_loc, load_func=np.load))

    data_files_series = pd.Series(data_files, dtype='file')
    assert data_files_series.dtype.name == 'file'

    data = Dataset([1, 2, 3], columns=['1'])
    data['data'] = data_files
    data['data2'] = data['data'].astype('file')
    assert data['data2'].dtype.name == 'file'

    data = Dataset([1, 2, 3], columns=['1'])
    data['data'] = data_files_series
    assert data['data'].dtype.name == 'file'

    # Clean up
    for i in range(3):
        file_loc = os.path.join(temp_dr, 'fake' + str(i) + '.npy')
        os.remove(file_loc)



import tempfile
import random
import numpy as np
import os
import inspect
from copy import deepcopy
from joblib import hash as joblib_hash
from ..util import is_array_like
from ..default.params.Params import Params


def to_memmap(X):

    f = os.path.join(tempfile.gettempdir(), str(random.random()))
    x = np.memmap(f, dtype=X.dtype, shape=X.shape, mode='w+')
    x[:] = X

    return f, X.dtype, X.shape


def from_memmap(X):

    X_file, X_type, X_shape = X
    X = np.memmap(X_file, dtype=X_type, shape=X_shape, mode='c')
    return X


def is_ng(p):

    try:
        return 'nevergrad' in p.__module__
    except AttributeError:
        return False


def get_grid_params(params):

    # Set grid params
    grid_params = {}
    for p in params:

        if hasattr(params[p], 'to_grid'):
            grid_params[p] = params[p].to_grid()
        elif is_ng(params[p]):
            raise RuntimeError('Passing nevergrad parameters directly is not '
                               'supported. Pass as a BPt Params equivalent.')
        else:
            grid_params[p] = params[p]

    return grid_params


def get_mean_fis(estimators, prop):

    fis = []

    # Go through each trained estimator
    for est in estimators:
        if hasattr(est, prop):
            fi = getattr(est, prop)

            # If any is None, return None
            if fi is None:
                return None

            # Try to squeeze values
            try:
                fi = fi.squeeze()
            except AttributeError:
                pass

            fis.append(fi)

        # If any don't, return None
        else:
            return None

    # Make sure all same len
    if len(set([len(x) for x in fis])) != 1:
        return None

    # Return as mean
    return np.mean(np.array(fis), axis=0)


def proc_mapping(indx, mapping):

    # Special case, if passed index of Ellipsis
    # return as is.
    if indx is Ellipsis:
        return indx

    # If non-empty mapping and non-empty index
    if len(mapping) > 0 and len(indx) > 0:

        # If should proc list...
        if is_array_like(indx[0]):
            return [proc_mapping(i, mapping) for i in indx]

        new_indx = set()

        for i in indx:
            new = mapping[i]

            if new is None:
                pass

            # If mapping points to a list of values
            elif isinstance(new, list):
                for n in new:
                    if n is not None:
                        new_indx.add(n)
            else:
                new_indx.add(new)

        # Sort, then return
        new_indx = sorted(list(new_indx))
        return new_indx

    # Base case return indx
    return indx

# @ TODO change lists in mapping interally to sets???


def update_mapping(mapping, new_mapping):

    # In case where new_mapping is empty,
    # return the original mapping as is
    if len(new_mapping) == 0:
        return

    # Go through the mapping and update each key with the new mapping
    for key in mapping:
        val = mapping[key]

        # In case of list
        if isinstance(val, list):

            # Generate a new list of values
            new_vals = []
            for v in val:

                # If the value is valid / in the new mapping
                if v in new_mapping:
                    new_val = new_mapping[v]

                    # Case where the new values is also a list
                    if isinstance(new_val, list):
                        new_vals += new_val
                    else:
                        new_vals.append(new_val)

                else:
                    new_vals.append(v)

            # Treat as set to remove duplicates
            as_set = set(new_vals)

            # Don't keep None's if any in list
            try:
                as_set.remove(None)
            except KeyError:
                pass

            # Exists case where None was the only element in the
            # list, in that case just set to None
            if len(as_set) == 0:
                mapping[key] = None

            # Otherwise set as list in sorted order
            else:
                mapping[key] = sorted(list(as_set))

        # Assume int if not list
        else:

            if val in new_mapping:
                mapping[key] = new_mapping[val]


def get_reverse_mapping(mapping):

    reverse_mapping = {}
    for m in mapping:
        key = mapping[m]

        if isinstance(key, list):
            for k in key:
                reverse_mapping[k] = m
        else:
            reverse_mapping[key] = m

    return reverse_mapping


def param_len_check(names, params, _print=print):

    if isinstance(params, dict) and len(names) == 1:
        return params

    try:

        if len(params) > len(names):
            _print('Warning! More params passed than objs')
            _print('Extra params have been truncated.')
            return params[:len(names)]

    # If non list params here
    except TypeError:
        return [0 for i in range(len(names))]

    while len(names) != len(params):
        params.append(0)

    return params


def replace_model_name(base_estimator_params):

    new = {}

    for key in base_estimator_params:
        value = base_estimator_params[key]

        split_key = key.split('__')
        split_key[0] = 'estimator'

        new_key = '__'.join(split_key)
        new[new_key] = value

    return new


def set_n_jobs(obj, n_jobs):

    # Call recursively for list
    if isinstance(obj, list) or isinstance(obj, tuple):
        for o in obj:
            set_n_jobs(o, n_jobs)

    # Check and set for n_jobs
    if hasattr(obj, 'n_jobs'):
        setattr(obj, 'n_jobs', n_jobs)


def get_possible_params(estimator, method):

    if not hasattr(estimator, method):
        return []

    pos_params = dict(inspect.getmembers(getattr(estimator, method).__code__))
    return pos_params['co_varnames']


def check_replace(objs):

    if isinstance(objs, list):
        return [check_replace(o) for o in objs]

    if isinstance(objs, set):
        new_set = set()
        for o in objs:
            new_set.add(check_replace(o))
        return new_set

    if isinstance(objs, tuple):
        return tuple([check_replace(o) for o in objs])

    if isinstance(objs, dict):
        return {k: check_replace(objs[k]) for k in objs}

    if hasattr(objs, 'get_params'):
        for param in objs.get_params(deep=False):
            new_value = check_replace(getattr(objs, param))
            setattr(objs, param, new_value)

        # Also if has n_jobs replace all with fixed 1
        if hasattr(objs, 'n_jobs'):
            setattr(objs, 'n_jobs', 1)
        if hasattr(objs, 'fix_n_jobs'):
            setattr(objs, 'fix_n_jobs', 1)
        if hasattr(objs, '_n_jobs'):
            try:
                setattr(objs, '_n_jobs', 1)
            except AttributeError:
                pass
        if hasattr(objs, 'n_jobs_'):
            try:
                setattr(objs, 'n_jobs_', 1)
            except AttributeError:
                pass

        # Return objs as changed in place
        return objs

    # If nevergrad / params convert to repr
    if isinstance(objs, Params):
        return repr(objs)

    # Return identity otherwise
    return objs


def pipe_hash(objs, steps):
    '''Expects a list'''

    # Make copy with nevergrad / params dists replaced by repr
    hash_steps = check_replace(deepcopy(steps))

    # Hash steps and objs separate, then combine
    hash_str1 = joblib_hash(objs, hash_name='md5')
    hash_str2 = joblib_hash(hash_steps, hash_name='md5')

    return hash_str1 + hash_str2


def replace_with_in_params(params, original, replace):

    new_params = {}

    for key in params:
        new_params[key.replace(original, replace)] = params[key]

    return new_params



from ..MLP import MLPClassifier_Wrapper, MLPRegressor_Wrapper
import numpy as np


def test_classifier():

    est = MLPClassifier_Wrapper(hidden_layer_sizes=[3.0, 3.0], tol=1)
    est.fit(np.ones((20, 20)), np.ones(20))


def test_regressor():

    est = MLPRegressor_Wrapper(hidden_layer_sizes=[3.0, 3.0], tol=1)
    est.fit(np.ones((20, 20)), np.ones(20))



import nevergrad as ng
from ..helpers import args_repr


class Params():

    def _class_name(self):
        return type(self).__name__

    def _extra_repr(self):

        if not hasattr(self, 'extra'):
            return ''

        base = ''
        for info in self.extra:
            base += '.' + info['name'] + '('
            base += args_repr(info['args'], info['kwargs'])
            base += ')'

        return base

    def __repr__(self):

        base = self._class_name() + '('
        base += args_repr(self.args_, self.kwargs_) + ')'
        return base + self._extra_repr()

    def __init__(self, *args, **kwargs):

        self.args_ = list(args)
        self.kwargs_ = kwargs

        super().__init__(*self.args_, **self.kwargs_)

    @property
    def value(self):
        '''Stores the value of the parameter.'''
        return super().value

    @value.setter
    def value(self, value):
        super().value = value

    @property
    def descriptors(self):
        '''Stores the descriptor information for this parameter.'''
        return super().descriptors

    @descriptors.setter
    def descriptors(self, descriptors):
        super().descriptors = descriptors


class ValParams(Params):

    def set_integer_casting(self):

        if not hasattr(self, 'extra'):
            self.extra = []

        info = {}
        info['name'] = 'set_integer_casting'
        info['args'] = []
        info['kwargs'] = {}
        self.extra.append(info)

        super().set_integer_casting()

        return self

    def set_mutation(self, *args, **kwargs):

        if not hasattr(self, 'extra'):
            self.extra = []

        info = {}
        info['name'] = 'set_mutation'
        info['args'] = list(args)
        info['kwargs'] = kwargs
        self.extra.append(info)

        super().set_mutation(*args, **kwargs)

        return self

    def set_bounds(self, *args, **kwargs):

        if not hasattr(self, 'extra'):
            self.extra = []

        info = {}
        info['name'] = 'set_bounds'
        info['args'] = list(args)
        info['kwargs'] = kwargs
        self.extra.append(info)

        super().set_bounds(*args, **kwargs)

        return self

    def set_recombination(self, *args, **kwargs):
        '''Sets a recombination mutation.

        For example:

        ::

            array.set_recombination(ng.p.mutation.Crossover(axis=(0, 1)))

        '''

        if not hasattr(self, 'extra'):
            self.extra = []

        info = {}
        info['name'] = 'set_recombination'
        info['args'] = list(args)
        info['kwargs'] = kwargs
        self.extra.append(info)

        super.set_recombination(*args, **kwargs)


# Override doc string
ValParams.set_integer_casting.__doc__ =\
    ng.p.Scalar.set_integer_casting.__doc__
ValParams.set_mutation.__doc__ =\
    ng.p.Scalar.set_mutation.__doc__
ValParams.set_bounds.__doc__ =\
    ng.p.Scalar.set_bounds.__doc__


def choice_to_grid(self):
    '''This method will attempt to convert from the
    current BPt / nevergrad style parameter to a sklearn-grid
    search compatible one.'''

    if 'choices' in self.kwargs_:
        choices = self.kwargs_['choices']
    else:
        choices = self.args_[0]

    as_grid = []
    for choice in choices:

        # Check for nested
        if hasattr(choice, 'to_grid'):
            values = choice.to_grid()
            if isinstance(values, list):
                as_grid += values
            else:
                as_grid.append(values)

        # If not a function with to_grid, add as is
        else:
            as_grid.append(choice)

    return as_grid


def undefined_to_grid(self):
    '''This parameter does not support converting to
    sklearn-grid compatible params.'''
    raise RuntimeError(self._class_name() + ' does not support GridSearch.')


class TransitionChoice(Params, ng.p.TransitionChoice):
    '''BPt parameter wrapper around :class:`nevergrad.p.TransitionChoice`.'''
    to_grid = choice_to_grid
    ChoiceTag = ng.p.TransitionChoice.ChoiceTag
    ChoiceTag.__doc__ = '''Used to identity param as choice.'''


class Choice(Params, ng.p.Choice):
    '''BPt parameter wrapper around :class:`nevergrad.p.Choice`.

    Examples
    ---------
    Constructing a Choice parameter:

    .. ipython:: python

        import BPt as bp
        class_weight_choice = bp.p.Choice(['balanced', None])
        class_weight_choice

    This parameter can then be set when setting :ref:`api.pipeline_pieces`.
    In the example below we will set it to be the params when
    constructing a logistic ridge model.

    .. ipython:: python

        params = {'class_weight': class_weight_choice}
        model = bp.Model('ridge', params=params)
        model

    '''
    to_grid = choice_to_grid
    ChoiceTag = ng.p.Choice.ChoiceTag
    ChoiceTag.__doc__ = '''Used to identity param as choice.'''


class Array(ValParams, ng.p.Array):
    '''BPt parameter wrapper around :class:`nevergrad.p.Array`.'''
    to_grid = undefined_to_grid


class Scalar(ValParams, ng.p.Scalar):
    '''BPt parameter wrapper around :class:`nevergrad.p.Scalar`.'''

    def to_grid(self):
        '''This method will attempt to convert from the
           current BPt / nevergrad style parameter to a sklearn-grid
           search compatible one.

           Since the base value is a scaler, a lower
           and upper bound must be set, and also this
           Scaler param must have been cast to an integer first.
           '''

        # If not interger
        if not self.integer:
            raise RuntimeError('Scalar cannot convert to grid unless cast '
                               'to integer.')

        lower = self.bounds[0]
        if len(lower) == 1:
            lower = int(lower[0])
        else:
            lower = None

        upper = self.bounds[1]
        if len(upper) == 1:
            upper = int(upper[0])
        else:
            upper = None

        if lower is not None and upper is not None:
            return list(range(lower, upper+1))

        raise RuntimeError('Scalar cannot convert to grid unless cast '
                           'both lower and upper are set')


class Log(ValParams, ng.p.Log):
    '''BPt parameter wrapper around :class:`nevergrad.p.Log`.'''
    to_grid = undefined_to_grid


class Tuple(Params, ng.p.Tuple):
    '''BPt parameter wrapper around :class:`nevergrad.p.Tuple`.'''
    to_grid = undefined_to_grid


class Instrumentation(Params, ng.p.Instrumentation):
    '''BPt parameter wrapper around :class:`nevergrad.p.Instrumentation`.'''
    to_grid = undefined_to_grid


class Dict(Params, ng.p.Dict):
    '''BPt parameter wrapper around :class:`nevergrad.p.Dict`.'''
    to_grid = undefined_to_grid


# @TODO Add more support for different to_grid?



import shutil
from .datasets import get_fake_dataset, get_fake_multi_index_dataset
import numpy as np
import tempfile
import os
import pytest


def test_add_data_files():

    df = get_fake_dataset()

    def file_to_subject(i):
        return int(i.split('_')[-1])

    files = {'q': ['a_0', 'b_1', 'c_2']}

    df = df.add_data_files(files=files,
                           file_to_subject=file_to_subject,
                           load_func=np.load)

    assert len(df['q']) == 3
    assert df.loc[2, 'q'] == 2

    assert 'a_0' in df.file_mapping[0].loc
    assert 'b_1' in df.file_mapping[1].loc
    assert 'c_2' in df.file_mapping[2].loc

    # Test in place
    df = get_fake_dataset()
    df.add_data_files(files=files,
                      file_to_subject=file_to_subject,
                      load_func=np.load, inplace=True)
    assert len(df['q']) == 3
    assert df.loc[2, 'q'] == 2
    assert 'a_0' in df.file_mapping[0].loc
    assert 'b_1' in df.file_mapping[1].loc
    assert 'c_2' in df.file_mapping[2].loc


def setup_datafiles():

    df = get_fake_dataset()

    def file_to_subject(i):
        return int(i.split('/')[-1].replace('.npy', ''))

    temp_dr = tempfile.gettempdir()
    locs = []
    for i in range(3):
        loc = os.path.join(temp_dr, str(i) + '.npy')
        locs.append(loc)

        data = np.zeros(shape=2)
        data[:] = i
        np.save(loc, data)

    files = {'data_files': locs}

    # Test add data files
    df = df.add_data_files(files=files,
                           file_to_subject=file_to_subject,
                           load_func=np.load)

    return df


def test_data_files_consolidate():

    df = setup_datafiles()
    assert df.file_mapping[0].load().shape == (2,)

    df['data_files2'] = df['data_files'].copy()
    df.add_scope('data_files2', 'data file', inplace=True)

    temp_dr = tempfile.gettempdir()
    save_dr = os.path.join(temp_dr, 'save_consol')

    if os.path.exists(save_dr):
        shutil.rmtree(save_dr)

    df.consolidate_data_files(save_dr, replace_with='test',
                              scope='data file', cast_to=None,
                              clear_existing='fail', n_jobs=-1)

    assert len(df.file_mapping) == 3
    assert df.shape == (3, 4)
    assert df.file_mapping[0].load().shape == ((2, 2))

    # Should fail if try again
    with pytest.raises(RuntimeError):
        df.consolidate_data_files(save_dr, replace_with='test',
                                  scope='data file', cast_to=None,
                                  clear_existing='fail', n_jobs=-1)


def test_data_files_consolidate2():

    df = setup_datafiles()
    assert df.file_mapping[0].load().shape == (2,)

    df['data_files2'] = df['data_files'].copy()
    df.add_scope('data_files2', 'data file', inplace=True)

    temp_dr = tempfile.gettempdir()
    save_dr = os.path.join(temp_dr, 'save_consol')

    if os.path.exists(save_dr):
        shutil.rmtree(save_dr)

    df.consolidate_data_files(save_dr, replace_with='test',
                              scope='data file', cast_to='float64',
                              clear_existing='fail', n_jobs=-1)

def test_data_files_integration():

    df = setup_datafiles()

    assert len(df['data_files']) == 3
    assert df.loc[0, 'data_files'] == 0
    assert df.loc[2, 'data_files'] == 2

    # Test get values
    values = df.get_values(col='data_files', reduce_func=np.mean)
    assert values.loc[0] == 0
    assert values.loc[1] == 1
    assert values.loc[2] == 2

    values = df.get_values(col='data_files', reduce_func=np.max, n_jobs=2)
    assert values.loc[0] == 0
    assert values.loc[1] == 1
    assert values.loc[2] == 2

    # Test auto detect categorical
    df = df.auto_detect_categorical()
    cat_cols = df.get_cols(scope='category')
    assert cat_cols == ['2', '3']

    # This also tests copy
    df_copy = df.copy(deep=True)

    # Drop outliers by std, should drop all by 1
    df_copy.filter_outliers_by_std(n_std=.5, scope='data file',
                                   drop=True, reduce_func=np.mean,
                                   n_jobs=1, inplace=True)

    # Make sure self._check_file_mapping works
    assert len(df_copy.file_mapping) == 1
    assert 1 in df_copy.file_mapping
    assert len(df_copy) == 1
    assert len(df.file_mapping) == 3

    # Try drop = False
    df_copy = df.copy(deep=True)
    df_copy = df_copy.filter_outliers_by_std(n_std=.5, scope='data file',
                                             drop=False, reduce_func=np.mean,
                                             n_jobs=1)

    assert np.nan in df_copy.file_mapping
    assert len(df_copy.file_mapping) == 2
    assert 1 in df_copy.file_mapping
    assert len(df_copy) == 3


def test_multi_index_add_data_files():

    df = get_fake_multi_index_dataset()

    def file_to_subject(i):

        subj = i.split('_')[1]
        event = i.split('_')[2]

        return (subj, event)

    # Leave c_s3_e2' as NaN
    files = {'files': ['a_s1_e1', 'a_s1_e2',
                       'b_s2_e1', 'b_s2_e2', 'c_s3_e1']}

    df = df.add_data_files(files=files,
                           file_to_subject=file_to_subject,
                           load_func=np.load)

    assert len(df['files']) == 6
    assert 'a_s1_e1' in df.file_mapping[0].loc


def test_to_data_file():

    df = get_fake_dataset()
    df['file'] = ['loc1', 'loc2', 'loc3']

    df.to_data_file('file', inplace=True)
    file_mapping = df.get_file_mapping()

    assert len(file_mapping) == 3
    assert 0 in file_mapping
    assert file_mapping[1].loc.endswith('loc2')


def test_to_data_file_not_inplace():

    df = get_fake_dataset()
    df['file'] = ['loc1', 'loc2', 'loc3']

    df = df.to_data_file('file', inplace=False)
    file_mapping = df.get_file_mapping()

    assert len(file_mapping) == 3
    assert 0 in file_mapping
    assert file_mapping[2].loc.endswith('loc3')



from .BPtEvaluator import BPtEvaluator
from .input_operations import BPtInputMixIn
from copy import deepcopy, copy
import pandas as pd
from .stats_helpers import  compute_corrected_ttest
import numpy as np
from itertools import combinations
from math import factorial


def clean_str(in_str):

    in_str = str(in_str)
    in_str = in_str.replace('"', '')
    in_str = in_str.replace("'", '')

    return in_str


def str_to_option(option):

    key, name = option.split('=')
    o = Option(value=None, name=name, key=key)
    return o


def str_to_options(option):

    # If single str option
    if option.count('=') == 1:
        return [str_to_option(option)]

    elif option.count('=') == 0:
        raise KeyError(repr(option) + 'is not a valid key.')

    # Otherwise multiple
    return [str_to_option(opt_str)
            for opt_str in option.split(', ')]


def add_scores(cols, evaluator, attr_name):

    if not hasattr(evaluator, attr_name):
        return

    attr = getattr(evaluator, attr_name)
    for key in attr:
        val = attr[key]

        try:
            cols[attr_name + '_' + key].append(val)
        except KeyError:
            cols[attr_name + '_' + key] = [val]


class Option(BPtInputMixIn):
    '''This is a special BPt input class designed
    to be used with :class:`Compare`. It is used as
    a formal way to represent comparison options.

    Parameters
    ----------
    value : option value
        The explicit value in which the parameter
        it is passed to will take.

    name : str or 'repr', optional
        The name of this value, where this name
        is used to index the :class:`CompareDict`
        of results.

        If left as default value of 'repr' then
        a str representation of the value will
        be automatically created.

        ::

            default = 'repr'

    key : str or None, optional
        This parameter should typically not be
        set directly. It is instead inferred by the
        context in which the option is eventually used.

        ::

            default = None

    '''

    def __init__(self, value, name='repr', key=None):

        self.value = value

        # The name is either by default
        # the representation or a custom name.
        if name == 'repr':

            if hasattr(self.value, 'obj') and \
             isinstance(getattr(self.value, 'obj'), str):
                self.name = clean_str(getattr(self.value, 'obj'))
            else:
                self.name = clean_str(repr(value))

        else:
            self.name = clean_str(name)

        if not isinstance(self.name, str):
            raise RuntimeError('name must be a str.')

        self.key = key

    def __hash__(self):
        return hash(self.name) + hash(self.key)

    def __repr__(self):

        if self.key is None:
            return self.name

        return self.key + '=' + self.name

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return (self.name, self.key) == (other.name, other.key)

    def __ne__(self, other):
        return not(self == other)

    # Sort by key then name if keys are equal
    def __lt__(self, other):

        if self.key == other.key:
            return self.name < other.name
        return self.key < other.key

    def __gt__(self, other):

        if self.key == other.key:
            return self.name > other.name
        return self.key > other.key


class Compare(BPtInputMixIn):
    '''This is a special BPt input class which
    can be used as a helper to more easily run comparison
    analysis between a few choice of parameters.

    Parameters
    ------------
    options : list of values or :class:`Option`
        This parameter should be a list of options
        in which to try each one of. You may pass
        these options either directly as values, for example ::

            options = ['option1', 'option2']

        Or as instances of :class:`Option`. This
        second strategy is reccomended when the underlying
        options are objects or something more complex then strings,
        for example between two :class:`Pipeline` ::

            pipe1 = bp.Pipeline([bp.Model('elastic')])
            pipe2 = bp.Pipeline([bp.Model('ridge')])

            options = [bp.Option(pipe1, name='pipe1'),
                       bp.Option(pipe2, name='pipe2')]


    Notes
    ---------
    | Usage of this object is designed to passed as input
        to :func:`evaluate`. Only parameters within :func:`evaluate`
        parameters `pipeline` and `problem_spec` (or their associated extra
        params) can be passed Compare. That said, some options, while valid
        may still make downstream intreptation more difficult, e.g., passing
        problem_type with two different Compare values will work, but
        will yield results with different metrics.

    | When to use Compare? It may be tempting to use Compare to evaluate
        different configurations of hyper-parameters, but in most cases
        this type of fine-grained usage is discouraged. On a conceptual level,
        the usage of Compare should be used to compare the actual underlying
        topic of interest! For example, if it
        of interest to the underlying research topic,
        then Compare can be used between two different :class:`Pipeline`.
        If instead this is not the key point of interest, but you still wish to
        try two different, say, :class:`Model`, then you would be better off
        nesting this choice as a hyper-parameter to optimize
        (in this case see: :class:`Select`).

    Examples
    ---------
    Compare is used with :func:`evaluate`, for example:

    .. ipython:: python

        from BPt.datasets import load_boston
        data = bp.datasets.load_boston()
        data.shape

        pipe_options = bp.Compare([bp.Option(bp.Model('elastic'),
                                             name='elastic'),
                                   bp.Option(bp.Model('ridge'),
                                             name='ridge')])

        bp.evaluate(pipe_options, data, progress_bar=False).summary()

    '''

    def __init__(self, options):

        self.options = []
        for option in options:

            # If not Option, cast to Option
            if not isinstance(option, Option):
                self.options.append(Option(option))

            # Otherwise, add as is
            else:
                self.options.append(option)

    def _check_args(self):
        '''If called here, then Compare is made
        up of pipeline pieces.'''

        for option in self.options:
            option.value._check_args()

    def _uniquify(self):

        for option in self.options:

            # Recursive check first
            if hasattr(option.value, '_uniquify'):
                option.value._uniquify()

            # Then replace with copy
            option.value = copy(option.value)


class Options():

    def __init__(self, *args, **kwargs):

        if len(args) == 1 and len(kwargs) == 0:
            self._single_arg(args[0])

        elif len(args) == 0 and len(kwargs) > 0:
            self._just_kwargs(kwargs)

    def _single_arg(self, first_arg):

        if isinstance(first_arg, list):

            options = []
            for option in first_arg:

                # skip if None
                if option is None:
                    continue

                # If option, add
                elif isinstance(option, Option):
                    options.append(option)

                # If options, add all existing options
                elif isinstance(option, Options):
                    options += option.options

                # Otherwise treat as query str and
                # cast to option.
                elif isinstance(option, str):
                    options += str_to_options(option)

                else:
                    raise RuntimeError('Unknown option:' + repr(option))

            self.options = options

        else:

            if isinstance(first_arg, str):

                try:
                    self.options = str_to_options(first_arg)

                except KeyError:
                    # If Failed, treat as Option with just value
                    self.options = [Option(value=None,
                                           name=first_arg,
                                           key='')]

            elif isinstance(first_arg, Option):
                self.options = [first_arg]
            else:
                self.options = first_arg

    def _just_kwargs(self, kwargs):

        self.options = [Option(value=None, name=repr(kwargs[key]), key=key)
                        for key in kwargs]

    def __hash__(self):
        return sum([hash(o) for o in self.options])

    def __repr__(self):
        return 'Options(' + ', '.join([repr(o) for
                                       o in sorted(self.options)]) + ')'

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return sorted(self.options) == sorted(other.options)

    def __ne__(self, other):
        return not(self == other)

    def is_subset(self, other):

        in_other = [option for option in self.options
                    if option in other.options]

        return len(in_other) == len(self.options)

    def is_just_name_subset(self, other):

        if len(self.options) > 1:
            return False

        name = self.options[0].name
        names = [option.name for option in other.options]

        if name in names:
            return True

        return False


class CompareDict(dict):
    '''This object is simmilar to a python dictionary,
    but is used to stored results from analyses run using
    :class:`Compare`.
    '''

    def _cast_to_options(self, k):

        # options can be None
        if k is None:
            return None

        if not isinstance(k, Options):
            k = Options(k)

        return k

    def __setitem__(self, k, v):

        k = self._cast_to_options(k)
        return super().__setitem__(k, v)

    def __getitem__(self, k):

        # Get input key as cast to Options
        k = self._cast_to_options(k)

        # First try to treat as single key
        try:
            return super().__getitem__(k)

        # If that fails, then check to see
        # if this index works as a subset.
        except KeyError:

            subset_keys = [key for key in self.keys() if k.is_subset(key)]

            # If fails here, try single subset
            if len(subset_keys) == 0:

                subset_keys = [key for key in self.keys()
                               if k.is_just_name_subset(key)]

                if len(subset_keys) == 0:
                    raise KeyError(repr(k) + ' is not a valid key.')

            # If just one, return that one
            if len(subset_keys) == 1:
                return self.__getitem__(subset_keys[0])

            # If any subsets, return a subset of this
            # dict with only those subsets
            return CompareDict({key: self[key] for key in subset_keys})

    def __repr__(self):
        return 'CompareDict(' + super().__repr__() + ')'

    def __str__(self):
        return self.__repr__()

    def summary(self):
        '''Return a pandas DataFrame with summary
        information as broken down by all of the different
        original :class:`Compare` options.'''

        # Get example value to base summary on
        ex = self.__getitem__(list(self.keys())[0])

        # @TODO make sure all of same time

        # if evaluation results
        if isinstance(ex, BPtEvaluator):
            return self._evaluator_summary()

        # @TODO add more options

        raise RuntimeError('Base options not comparable.')

    def _evaluator_summary(self):

        keys = list(self.keys())
        repr_key = keys[0]
        option_keys = [o.key for o in repr_key.options]
        cols = {key: [] for key in option_keys}

        for key in list(self.keys()):

            for o in key.options:
                cols[o.key].append(o.name)

            # Add values
            evaluator = self[key]
            add_scores(cols, evaluator, attr_name='mean_scores')
            add_scores(cols, evaluator, attr_name='std_scores')
            add_scores(cols, evaluator, attr_name='mean_timing')

        summary = pd.DataFrame.from_dict(cols)
        summary = summary.set_index(option_keys)

        return summary

    def pairwise_t_stats(self, metric='first'):
        '''This method performs pair-wise t-test
        comparisons between all different options,
        assuming this object holds instances of :class:`BPtEvaluator`.
        The method used to generate t-test comparisons here is based off the
        example code from:
        https://scikit-learn.org/stable/auto_examples/model_selection/plot_grid_search_stats.html

        .. note::
            In the case that the sizes of the training and validation sets
            at each fold vary dramatically, it is unclear if this
            statistics are still valid.
            In that case, the mean train size and mean validation sizes
            are employed when computing statistics.

        Parameters
        ----------
        metric : str, optional
            This method compares the metrics
            produced for only one valid
            metric / scorer. Notably
            all :class:`BPtEvaluator` must have
            been evaluated with respect to this
            scorer. By default the reserved key, 'first'
            indicates that just whatever scorer is first
            should be used to produce the pairwise
            t statistics.

            ::

                default = 'first'

        Returns
        --------
        stats_df : pandas DataFrame
            A DataFrame comparing all pairwise
            combinations of the original :class:`Compare` options.
            't_stat' and 'p_val' columns will be generated for
            each comparison representing the corrected t_stat
            for non-independence of folds and the corresponding
            Bonferroni correctted p values (for multiple comparisons
            from comparing all pairwise combinations). See the
            referenced scikit-learn example for more information.
        '''

        # @TODO Clean up code and add error handling.

        ex_data_point = self.__getitem__(list(self.keys())[0])
        metrics = list(ex_data_point.mean_scores)

        # If metric is first
        if metric == 'first':
            metric = metrics[0]

        # Make sure valid metric
        if metric not in metrics:
            raise RuntimeError(f'{metric} not in avaliable {metrics}')

        n_comparisons = (
            factorial(len(self))
            / (factorial(2) * factorial(len(self) - 2))
        )

        pairwise_t_test = []
        for model_i, model_k in combinations(self, 2):

            # Grab scores
            model_i_scores = np.array(self[model_i].scores[metric])
            model_k_scores = np.array(self[model_k].scores[metric])

            # Skip if equal
            if np.array_equal(model_i_scores, model_k_scores):
                continue

            # Compute differences
            differences = model_i_scores - model_k_scores

            n = len(model_i_scores)
            df = n - 1

            # Use the mean train / test size
            n_train = np.mean([len(ti) for ti in self[model_i].train_indices])
            n_test = np.mean([len(ti) for ti in self[model_i].val_indices])

            # Do t-test
            t_stat, p_val =\
                compute_corrected_ttest(differences, df, n_train, n_test)

            # Implement Bonferroni correction
            p_val *= n_comparisons

            # Bonferroni can output p-values higher than 1
            p_val = 1 if p_val > 1 else p_val

            # Gen key entry names
            mi_option_names = [o.name for o in model_i.options]
            mk_option_names = [o.name for o in model_k.options]
            opt_names = mi_option_names + mk_option_names

            # Append
            pairwise_t_test.append(opt_names + [t_stat, p_val])

        ex = list(self)[0]
        key_names = [str(o.key) for o in ex.options]
        col_names = [key + ' (1)' for key in key_names] +\
            [key + ' (2)' for key in key_names]

        pairwise_comp_df = pd.DataFrame(
            pairwise_t_test, columns=col_names + ['t_stat', 'p_val'])

        return pairwise_comp_df

    def pairwise_bayesian():
        '''This method has not yet been implemented'''

        '''
        pairwise_bayesian = []

        for model_i, model_k in combinations(range(len(model_scores)), 2):
            model_i_scores = model_scores.iloc[model_i].values
            model_k_scores = model_scores.iloc[model_k].values
            differences = model_i_scores - model_k_scores
            t_post = t(
                df, loc=np.mean(differences),
                scale=corrected_std(differences, n_train, n_test)
            )
            worse_prob = t_post.cdf(rope_interval[0])
            better_prob = 1 - t_post.cdf(rope_interval[1])
            rope_prob = t_post.cdf(rope_interval[1]) - t_post.cdf(rope_interval[0])

            pairwise_bayesian.append([worse_prob, better_prob, rope_prob])

        pairwise_bayesian_df = (pd.DataFrame(
            pairwise_bayesian,
            columns=['worse_prob', 'better_prob', 'rope_prob']
        ).round(3))

        pairwise_comp_df = pairwise_comp_df.join(pairwise_bayesian_df)
        pairwise_comp_df
        '''

        return


def _make_compare_copies(objs, key, compare):

    # If not already compare dict, cast to compare dict
    # under empty string to represent base
    if not isinstance(objs, CompareDict):
        objs = CompareDict({None: objs})

    # Store all new objects
    new_objs = CompareDict()

    # For each compare value
    for option in compare.options:

        # Set key in option
        option.key = key

        # For each existing object
        for obj_key in objs:

            # Get copy
            new_obj = deepcopy(objs[obj_key])

            # Assign value in place
            new_obj.set_params(**{key: option.value})

            # New key as existing key + new option
            new_obj_key = Options([obj_key, option])

            # Add to new objs under new key
            if new_obj_key in new_objs:
                raise RuntimeError('key collision!')

            new_objs[new_obj_key] = new_obj

    return new_objs


def _compare_check(obj):

    # Check for any params set to Compare
    params = obj.get_params(deep=True)

    for key in params:
        if isinstance(params[key], Compare):

            # Update obj to be list of problem
            # specs init'ed with different options
            obj = _make_compare_copies(obj, key, params[key])

    return obj


def _merge_compare(pipe, ps):

    # If both are not dicts
    if not isinstance(pipe, CompareDict) and not isinstance(ps, CompareDict):
        return pipe, ps

    # New will stored new dict
    new = CompareDict()

    # If both are dicts
    if isinstance(pipe, CompareDict) and isinstance(ps, CompareDict):

        # Save each combined combination
        for pipe_key in pipe:
            for ps_key in ps:
                new[[pipe_key, ps_key]] =\
                    (deepcopy(pipe[pipe_key]), deepcopy(ps[ps_key]))

        return new

    # If just ps
    elif isinstance(ps, CompareDict):
        for ps_key in ps:
            new[ps_key] = (deepcopy(pipe), deepcopy(ps[ps_key]))

        return new

    # If just pipe
    for pipe_key in pipe:
        new[pipe_key] = (deepcopy(pipe[pipe_key]), deepcopy(ps))

    return new



from ..helpers import get_obj_and_params, all_from_objects

from sklearn.decomposition import (PCA, FactorAnalysis,
                                   MiniBatchDictionaryLearning,
                                   DictionaryLearning, FastICA,
                                   IncrementalPCA, KernelPCA,
                                   MiniBatchSparsePCA, NMF, SparsePCA,
                                   TruncatedSVD)

from sklearn.preprocessing import OneHotEncoder


TRANSFORMERS = {
    'pca': (PCA, ['default', 'pca var search']),
    'sparse pca': (SparsePCA, ['default']),
    'mini batch sparse pca': (MiniBatchSparsePCA, ['default']),
    'factor analysis': (FactorAnalysis, ['default']),
    'dictionary learning': (DictionaryLearning, ['default']),
    'mini batch dictionary learning': (MiniBatchDictionaryLearning,
                                       ['default']),
    'fast ica': (FastICA, ['default']),
    'incremental pca': (IncrementalPCA, ['default']),
    'kernel pca': (KernelPCA, ['default']),
    'nmf': (NMF, ['default']),
    'truncated svd': (TruncatedSVD, ['default']),
    'one hot encoder': (OneHotEncoder, ['ohe']),
    'dummy coder': (OneHotEncoder, ['dummy code'])}


def get_transformer_and_params(transformer_str, extra_params,
                               params, **kwargs):

    transformer, extra_transformer_params, transformer_params =\
        get_obj_and_params(transformer_str, TRANSFORMERS, extra_params, params)

    return transformer(**extra_transformer_params), transformer_params


all_obj_keys = all_from_objects(TRANSFORMERS)



from BPt.pipeline.ensemble_wrappers import BPtVotingRegressor
from ..BPtModel import BPtModel
from ..constructors import (add_estimator_to_params, FeatSelectorConstructor,
                            ScalerConstructor, ModelConstructor)
from ...main.input import Ensemble, FeatSelector, Model, Scaler
from ..BPtFeatureSelector import BPtFeatureSelector
from .helpers import get_fake_data_dataset
from sklearn.ensemble import RandomForestRegressor
from sklearn.compose import TransformedTargetRegressor
from sklearn.ensemble import AdaBoostRegressor


def test_add_estimator_to_params():

    params = {'s__1': 1, 'my_model__2': 2}
    returned = add_estimator_to_params(params)

    assert returned['s__estimator__1'] == 1
    assert returned['my_model__estimator__2'] == 2


def test_feature_selectors():

    data_keys = ['0', '1']
    dataset = get_fake_data_dataset(data_keys=data_keys)

    spec = {'problem_type': 'binary',
            'random_state': None,
            'n_jobs': 1,
            'scope': 'all'}

    fs = FeatSelectorConstructor(dataset=dataset, spec=spec,
                                 user_passed_objs={})

    in_params = FeatSelector('univariate selection')
    objs, params = fs.process(in_params)
    name, obj = objs[0]

    assert 'univariate selection' in name
    assert obj.inds is Ellipsis
    assert len(params) == 0
    assert isinstance(obj, BPtFeatureSelector)

    in_params = FeatSelector('univariate selection', params=1)
    objs, params = fs.process(in_params)
    name, obj = objs[0]

    assert 'univariate selection' in name
    assert obj.inds is Ellipsis
    assert len(params) > 0
    assert isinstance(obj, BPtFeatureSelector)

    for key in params:
        assert 'univariate selection' in key

    in_params = FeatSelector('univariate selection', scope='category')
    objs, params = fs.process(in_params)
    name, obj = objs[0]

    assert 'univariate selection' in name
    assert len(obj.inds) == 0
    assert len(params) == 0
    assert isinstance(obj, BPtFeatureSelector)

    dataset = get_fake_data_dataset(data_keys=data_keys, cat_keys=['1'])
    fs = FeatSelectorConstructor(dataset=dataset, spec=spec,
                                 user_passed_objs={})

    in_params = FeatSelector('univariate selection', scope='category')
    objs, params = fs.process(in_params)
    name, obj = objs[0]

    assert 'univariate selection' in name
    assert obj.inds == [1]
    assert len(params) == 0
    assert isinstance(obj, BPtFeatureSelector)


def test_feature_selectors_submodel():

    data_keys = ['0', '1']
    dataset = get_fake_data_dataset(data_keys=data_keys)
    spec = {'problem_type': 'binary', 'random_state': None,
            'n_jobs': 1, 'scope': 'all'}

    fs = FeatSelectorConstructor(dataset=dataset, spec=spec,
                                 user_passed_objs={})

    in_params = FeatSelector('rfe')
    objs, params = fs.process(in_params)
    name, obj = objs[0]

    assert 'rfe' in name
    assert len(params) == 0
    assert obj.inds is Ellipsis
    assert obj.estimator.estimator is None
    assert isinstance(obj, BPtFeatureSelector)

    in_params = FeatSelector(obj='rfe',
                             base_model=Model(obj='ridge', params=1))
    objs, params = fs.process(in_params)
    name, obj = objs[0]

    assert 'rfe' in name
    assert obj.inds is Ellipsis
    assert obj.estimator.estimator is not None
    assert len(params) > 1
    assert isinstance(obj, BPtFeatureSelector)

    # Each param should start with name estimator, estimator
    for key in params:
        assert key.startswith('__'.join([name, 'estimator', 'estimator']))


def test_empty_scope():

    data_keys = ['0', '1']
    dataset = get_fake_data_dataset(data_keys=data_keys)
    spec = {'problem_type': 'binary', 'random_state': None,
            'n_jobs': 1, 'scope': 'all'}

    sc = ScalerConstructor(dataset=dataset, spec=spec,
                           user_passed_objs={})

    in_params = Scaler('robust', scope='category')
    objs, params = sc.process(in_params)
    _, obj = objs[0]

    assert len(obj.inds) == 0
    assert len(params) == 0


def test_model():

    data_keys = ['0', '1']
    dataset = get_fake_data_dataset(data_keys=data_keys)

    spec = {'problem_type': 'regression', 'random_state': None,
            'n_jobs': 1, 'scope': 'all'}

    mc = ModelConstructor(dataset=dataset, spec=spec, user_passed_objs={})
    in_params = Model('random forest')

    objs, params = mc.process(in_params)
    _, obj = objs[0]
    assert isinstance(obj, BPtModel)
    assert isinstance(obj.estimator, RandomForestRegressor)

    assert obj.inds is Ellipsis
    assert len(params) == 0


def test_model_target_scaler():

    data_keys = ['0', '1']
    dataset = get_fake_data_dataset(data_keys=data_keys)

    spec = {'problem_type': 'regression', 'random_state': None,
            'n_jobs': 1, 'scope': 'all'}

    mc = ModelConstructor(dataset=dataset, spec=spec, user_passed_objs={})
    in_params = Model('random forest', scope='float',
                      target_scaler=Scaler('standard'))

    objs, params = mc.process(in_params)
    _, obj = objs[0]

    assert isinstance(obj, BPtModel)
    assert isinstance(obj.estimator, TransformedTargetRegressor)
    assert isinstance(obj.estimator.regressor, RandomForestRegressor)
    assert obj.inds is Ellipsis

    assert len(params) == 0


def test_ensemble_stacking():

    data_keys = ['0', '1']
    dataset = get_fake_data_dataset(data_keys=data_keys)

    spec = {'problem_type': 'regression', 'random_state': None,
            'n_jobs': 2, 'scope': 'all'}

    mc = ModelConstructor(dataset=dataset, spec=spec, user_passed_objs={})
    in_params = Ensemble('voting', [Model('random forest'),
                                    Model('random forest')],
                         n_jobs_type='ensemble')

    objs, params = mc.process(in_params)
    _, obj = objs[0]

    assert isinstance(obj, BPtModel)
    assert obj.inds is Ellipsis
    assert isinstance(obj.estimator, BPtVotingRegressor)
    assert obj.estimator.n_jobs == 2

    ests = obj.estimator.estimators
    for est in ests:
        assert isinstance(est[1], BPtModel)
        assert est[1].estimator.n_jobs == 1

    assert len(params) == 0


def test_ensemble_nested_stacking():

    data_keys = ['0', '1']
    dataset = get_fake_data_dataset(data_keys=data_keys)

    spec = {'problem_type': 'regression', 'random_state': None,
            'n_jobs': 2, 'scope': 'all'}

    mc = ModelConstructor(dataset=dataset, spec=spec, user_passed_objs={})
    in_params = Ensemble('voting', [Model('random forest'),
                                    Model('random forest'),
                                    Ensemble('voting', [Model('random forest'),
                                                        Model('random forest')])
                                    ],
                         n_jobs_type='ensemble')

    objs, params = mc.process(in_params)
    _, obj = objs[0]

    assert isinstance(obj, BPtModel)
    assert obj.inds is Ellipsis
    assert isinstance(obj.estimator, BPtVotingRegressor)
    assert obj.estimator.n_jobs == 2

    estimators = obj.estimator.estimators
    assert isinstance(estimators[1][1], BPtModel)
    assert isinstance(estimators[2][1], BPtModel)

    nested_ensemble = estimators[0][1].estimator
    assert isinstance(nested_ensemble, BPtVotingRegressor)
    assert nested_ensemble.n_jobs == 1

    nested_ests = nested_ensemble.estimators
    assert isinstance(nested_ests[0][1], BPtModel)
    assert isinstance(nested_ests[0][1].estimator, RandomForestRegressor)
    assert isinstance(nested_ests[1][1], BPtModel)
    assert isinstance(nested_ests[1][1].estimator, RandomForestRegressor)

    assert len(params) == 0


def test_ensemble_se():

    data_keys = ['0', '1']
    dataset = get_fake_data_dataset(data_keys=data_keys)

    spec = {'problem_type': 'regression', 'random_state': None,
            'n_jobs': 2, 'scope': 'all'}

    mc = ModelConstructor(dataset=dataset, spec=spec, user_passed_objs={})
    in_params = Ensemble(obj='adaboost',
                         models=Model('random forest'),
                         single_estimator=True)

    objs, params = mc.process(in_params)
    _, obj = objs[0]

    assert isinstance(obj, BPtModel)
    assert isinstance(obj.estimator, AdaBoostRegressor)
    assert len(params) == 0



from BPt.pipeline.BPtPipeline import BPtPipeline
from BPt.main.BPtEvaluator import BPtEvaluator
from ...pipeline.BPtSearchCV import BPtGridSearchCV, NevergradSearchCV
from ..input import (Loader, ModelPipeline, Model, CV, Pipeline, Scaler,
                     ParamSearch, FeatSelector, Transformer)
from ..input_operations import Select
from ...dataset.Dataset import Dataset
from ...default.params.Params import Choice
from ..funcs import evaluate, cross_val_score, _sk_check_y
import pytest
import pandas as pd
import numpy as np
from ...extensions import LinearResidualizer
from ..compare import Compare, Option, CompareDict, Options
from sklearn.tree import DecisionTreeClassifier
from ...pipeline.BPtLoader import BPtLoader
import warnings
from ...pipeline.Selector import Selector
from ...pipeline.tests.helpers import get_fake_mapping, clean_fake_mapping


def get_fake_datafile_dataset(m, n):

    fm = get_fake_mapping(int(m*n))
    fm = {key: fm[key].loc for key in fm}
    data = Dataset(np.arange(int(m*n)).reshape((m, n)),
                   columns=['col1', 'col2'],
                   dtype='object')
    data = data.replace(fm)

    data.to_data_file(scope='all', inplace=True)

    data['t'] = np.random.random(m)
    data.set_role('t', 'target', inplace=True)

    return data


def test_evaluate_with_loader():

    data = get_fake_datafile_dataset(50, 2)
    pipe = Pipeline([Loader('identity'), Model('dt')])

    results = evaluate(pipeline=pipe,
                       dataset=data,
                       progress_bar=False,
                       cv=3)

    assert len(results.feat_names[0]) == 8
    assert isinstance(results.estimators[0].steps[0][1], BPtLoader)

    # Clean up
    clean_fake_mapping(100)


def test_evaluate_with_loader_inverse_fis():

    data = get_fake_datafile_dataset(50, 2)
    pipe = Pipeline([Loader('identity'), Model('dt')])

    results = evaluate(pipeline=pipe,
                       dataset=data,
                       progress_bar=False,
                       cv=3)

    # Clean up
    clean_fake_mapping(100)

    inverse_fis = results.get_inverse_fis()

    assert len(inverse_fis) == 3
    assert inverse_fis[0].loc['col1'].shape == (2, 2)
    assert inverse_fis[0].loc['col2'].shape == (2, 2)


def test_evaluate_with_loader_inverse_fis_categorical():

    warnings.filterwarnings("ignore")

    data = get_fake_datafile_dataset(50, 2)
    pipe = Pipeline([Loader('identity'), Model('linear', tol=1)])

    data = data.k_bin(scope='t', n_bins=5)

    results = evaluate(pipeline=pipe,
                       dataset=data,
                       progress_bar=False,
                       cv=3,
                       random_state=3)
    # Clean up
    clean_fake_mapping(100)

    inverse_fis = results.get_inverse_fis()
    assert len(inverse_fis) == 3

    fold1 = inverse_fis[0]

    # One for each class
    assert len(fold1) == 5

    assert fold1[0].loc['col1'].shape == (2, 2)
    assert fold1[3].loc['col1'].shape == (2, 2)
    assert fold1[0].loc['col2'].shape == (2, 2)


def get_fake_dataset():

    fake = Dataset()

    fake['1'] = np.ones((20))
    fake['2'] = np.ones((20))
    fake['3'] = np.ones((20))
    fake = fake.set_role('3', 'target')

    return fake


def test_sk_check_y():

    # Make sure passes
    y = pd.Series([1, 2, 3])
    _sk_check_y(y)

    # Fails
    y = pd.Series([1, 2, np.nan])
    with pytest.raises(RuntimeError):
        _sk_check_y(y)


def test_evaluate_match_cross_val_score():

    dt_pipe = ModelPipeline(model=Model('dt'))
    dataset = get_fake_dataset()
    evaluator = evaluate(pipeline=dt_pipe,
                         dataset=dataset,
                         problem_spec='default',
                         cv=5,
                         scorer='r2',
                         progress_bar=False,
                         store_preds=True,
                         store_estimators=True,
                         store_timing=True,
                         progress_loc=None,
                         problem_type='regression')

    cv_scores = cross_val_score(pipeline=dt_pipe,
                                dataset=dataset,
                                scorer='r2',
                                problem_spec='default',
                                problem_type='regression')

    assert np.sum(evaluator.scores['r2']) == np.sum(cv_scores)


def test_evaluate_regression_dt():

    dt_pipe = ModelPipeline(model=Model('dt'))
    dataset = get_fake_dataset()
    evaluator = evaluate(pipeline=dt_pipe,
                         dataset=dataset,
                         problem_spec='default',
                         cv=5,
                         progress_bar=False,
                         store_preds=True,
                         store_estimators=True,
                         store_timing=True,
                         progress_loc=None,
                         problem_type='regression')

    assert len(evaluator.scores) > 0
    assert len(evaluator.timing['fit']) > 0
    assert len(evaluator.timing['score']) > 0
    evaluator._estimators_check()

    fis_df = evaluator.get_fis()
    assert list(fis_df) == ['1', '2']
    assert len(fis_df) == 5
    assert len(evaluator.feature_importances_) == 2

    raw_fis = evaluator.get_feature_importances()
    assert len(raw_fis) == 5
    assert evaluator.coef_ is None

    assert isinstance(evaluator.preds, dict)
    assert len(evaluator.preds['predict']) == 5
    assert len(evaluator.preds['y_true']) == 5

    assert len(evaluator.train_indices) == 5
    assert len(evaluator.val_indices) == 5


def test_regression_mean_fis():

    dt_pipe = ModelPipeline(model=Model('dt'))
    dataset = get_fake_dataset()
    evaluator = evaluate(pipeline=dt_pipe,
                         dataset=dataset,
                         progress_bar=False,
                         problem_type='regression',
                         random_state=0)

    mean_fis = evaluator.get_fis(mean=True)
    assert len(mean_fis) == 0


def get_fake_category_dataset():

    fake = Dataset()

    base = np.ones((30))
    base[0:10] = 0
    base[10:20] = 2

    fake['1'] = base
    fake['2'] = base
    fake['3'] = base

    fake = fake.set_role('3', 'target')

    return fake


def test_evaluate_categorical_dt():

    dt_pipe = ModelPipeline(model=Model('dt'))
    dataset = get_fake_category_dataset()

    evaluator = evaluate(pipeline=dt_pipe,
                         dataset=dataset,
                         problem_spec='default',
                         cv=5,
                         progress_bar=False,
                         store_preds=True,
                         store_estimators=True,
                         store_timing=True,
                         progress_loc=None,
                         problem_type='categorical')

    assert len(evaluator.scores) > 0
    assert len(evaluator.timing['fit']) > 0
    assert len(evaluator.timing['score']) > 0
    evaluator._estimators_check()

    fis_df = evaluator.get_fis()
    assert list(fis_df) == ['1', '2']
    assert len(fis_df) == 5
    assert len(evaluator.feature_importances_) == 2

    raw_fis = evaluator.get_feature_importances()
    assert len(raw_fis) == 5
    assert evaluator.coef_ is None

    predict = evaluator.preds['predict']
    predict_proba = evaluator.preds['predict_proba']
    y_true = evaluator.preds['y_true']

    assert len(predict) == 5
    assert predict[0].shape == y_true[0].shape
    assert predict[0].shape[0] == predict_proba[0].shape[0]


def test_evaluate_categorical_dt_mean_fis():

    dt_pipe = ModelPipeline(model=Model('linear'))
    dataset = get_fake_category_dataset()

    evaluator = evaluate(pipeline=dt_pipe,
                         dataset=dataset,
                         problem_spec='default',
                         progress_bar=False,
                         problem_type='categorical',
                         random_state=1,
                         cv=3)
    mean_fis = evaluator.get_fis(mean=True)
    assert len(mean_fis) == 3


def test_evaluate_categorical_linear():

    linear_pipe = ModelPipeline(model=Model('linear'))
    dataset = get_fake_category_dataset()

    evaluator = evaluate(pipeline=linear_pipe,
                         dataset=dataset,
                         problem_spec='default',
                         cv=5,
                         progress_bar=False,
                         store_preds=True,
                         store_estimators=True,
                         store_timing=True,
                         problem_type='categorical')

    assert len(evaluator.coef_) == len(evaluator.fis_)
    fis = evaluator.get_fis()
    first_fi = fis[0]
    assert isinstance(first_fi, pd.DataFrame)
    assert first_fi.shape == (5, 2)
    assert evaluator.feature_importances_ is None


def get_fake_binary_dataset():

    fake = Dataset()

    base = np.ones((20))
    base[:10] = 0

    fake['1'] = base
    fake['2'] = base
    fake['3'] = base

    fake = fake.set_role('3', 'target')

    return fake


def test_evaluate_binary():

    dt_pipe = ModelPipeline(model=Model('dt'))
    dataset = get_fake_binary_dataset()
    evaluator = evaluate(pipeline=dt_pipe,
                         dataset=dataset,
                         problem_spec='default',
                         cv=5,
                         progress_bar=False,
                         store_preds=True,
                         store_estimators=True,
                         store_timing=True,
                         progress_loc=None,
                         problem_type='binary')

    assert len(evaluator.scores) > 0
    assert len(evaluator.timing['fit']) > 0
    assert len(evaluator.timing['score']) > 0
    evaluator._estimators_check()

    fis_df = evaluator.get_fis()
    assert list(fis_df) == ['1', '2']
    assert len(fis_df) == 5
    assert len(evaluator.feature_importances_) == 2

    raw_fis = evaluator.get_feature_importances()
    assert len(raw_fis) == 5
    assert evaluator.coef_ is None

    predict = evaluator.preds['predict']
    predict_proba = evaluator.preds['predict_proba']
    y_true = evaluator.preds['y_true']

    assert len(predict) == 5
    assert predict[0].shape == y_true[0].shape
    assert predict[0].shape[0] == predict_proba[0].shape[0]


def test_evaluate_fail():

    dt_pipe = ModelPipeline(model=Model('dt'))
    dataset = get_fake_dataset()

    evaluator = evaluate(pipeline=dt_pipe,
                         dataset=dataset,
                         progress_bar=False,
                         store_estimators=False)

    with pytest.raises(RuntimeError):
        evaluator._estimators_check()

    with pytest.raises(RuntimeError):
        evaluator.permutation_importance(dataset)


def test_evaluate_cv():

    dt_pipe = ModelPipeline(model=Model('dt'))
    dataset = get_fake_binary_dataset()
    cv = CV(splits=3, n_repeats=2)
    evaluator = evaluate(pipeline=dt_pipe,
                         dataset=dataset,
                         problem_spec='default',
                         cv=cv,
                         progress_bar=False,
                         store_preds=True,
                         store_estimators=True,
                         store_timing=True,
                         progress_loc=None,
                         problem_type='binary')

    assert evaluator.n_repeats_ == 2
    assert evaluator.n_splits_ == 3
    assert len(evaluator.mean_scores) == 3
    assert len(evaluator.std_scores) == 6


def test_nan_targets():

    dt_pipe = ModelPipeline(model=Model('dt'))
    dataset = get_fake_binary_dataset()
    dataset.loc[0, '3'] = np.nan

    cv = CV(splits=3, n_repeats=1)
    evaluator = evaluate(pipeline=dt_pipe,
                         dataset=dataset,
                         cv=cv,
                         scorer='roc_auc',
                         progress_bar=False,
                         problem_type='binary')

    assert evaluator.n_repeats_ == 1
    assert evaluator.n_splits_ == 3

    for tr in evaluator.train_indices:
        assert 0 not in tr

    # Should be in one here
    is_in = [0 in tr for tr in evaluator.all_train_indices]
    assert any(is_in)

    assert len(evaluator.all_train_indices) == len(evaluator.train_indices)

    ati_len = len(sum([list(e) for e in evaluator.all_train_indices], []))
    ti_len = len(sum([list(e) for e in evaluator.train_indices], []))
    assert ati_len != ti_len

    avi_len = len(sum([list(e) for e in evaluator.all_val_indices], []))
    vi_len = len(sum([list(e) for e in evaluator.val_indices], []))
    assert avi_len != vi_len

    preds_dfs_dropt = evaluator.get_preds_dfs(drop_nan_targets=True)
    preds_dfs = evaluator.get_preds_dfs(drop_nan_targets=False)
    assert preds_dfs_dropt[0].shape != preds_dfs[0].shape


def test_only_fold_cv_param():

    dt_pipe = ModelPipeline(model=Model('dt'))
    dataset = get_fake_binary_dataset()
    cv = CV(splits=3, n_repeats=1, only_fold=0)
    evaluator = evaluate(pipeline=dt_pipe,
                         dataset=dataset,
                         cv=cv,
                         scorer='roc_auc',
                         progress_bar=False,
                         problem_type='binary')

    assert evaluator.n_repeats_ == 1
    assert evaluator.n_splits_ == 1
    assert len(evaluator.scores['roc_auc']) == 1
    assert len(evaluator.train_indices) == 1


def test_multiple_only_fold_cv_param():

    dt_pipe = ModelPipeline(model=Model('dt'))
    dataset = get_fake_binary_dataset()
    cv = CV(splits=3, n_repeats=2, only_fold=[0, 4])
    evaluator = evaluate(pipeline=dt_pipe,
                         dataset=dataset,
                         cv=cv,
                         scorer='roc_auc',
                         progress_bar=False,
                         problem_type='binary')

    assert evaluator.n_repeats_ == 1
    assert evaluator.n_splits_ == 2
    assert len(evaluator.scores['roc_auc']) == 2
    assert len(evaluator.train_indices) == 2


def test_evaluate_with_resid():

    dataset = get_fake_dataset()
    resid = LinearResidualizer(to_resid_df=dataset[['1']])
    pipe = Pipeline(steps=[Scaler(obj=resid, scope='all'), Model('dt')])

    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         problem_spec='default',
                         cv=5,
                         scorer='r2',
                         progress_bar=False,
                         store_preds=True,
                         store_estimators=True,
                         store_timing=True,
                         progress_loc=None,
                         problem_type='regression')

    # Test matches / works with cross_val_score
    cv_scores = cross_val_score(pipeline=pipe,
                                dataset=dataset,
                                scorer='r2',
                                problem_spec='default',
                                problem_type='regression')
    assert np.sum(evaluator.scores['r2']) == np.sum(cv_scores)

    assert len(evaluator.scores) > 0
    assert len(evaluator.timing['fit']) > 0
    assert len(evaluator.timing['score']) > 0
    evaluator._estimators_check()

    fis_df = evaluator.get_fis()
    assert list(fis_df) == ['1', '2']
    assert len(fis_df) == 5
    assert len(evaluator.feature_importances_) == 2

    raw_fis = evaluator.get_feature_importances()
    assert len(raw_fis) == 5
    assert evaluator.coef_ is None

    assert isinstance(evaluator.preds, dict)
    assert len(evaluator.preds['predict']) == 5
    assert len(evaluator.preds['y_true']) == 5

    assert len(evaluator.train_indices) == 5
    assert len(evaluator.val_indices) == 5


def test_evaluate_with_resid_param_search():

    dataset = get_fake_dataset()
    resid = LinearResidualizer(to_resid_df=dataset[['1']])
    pipe = Pipeline(steps=[Scaler(obj=resid, scope='all'),
                           Model('dt', params=1)],
                    param_search=ParamSearch(n_iter=5))

    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         problem_spec='default',
                         cv=5,
                         scorer='r2',
                         progress_bar=False,
                         store_preds=True,
                         store_estimators=True,
                         store_timing=True,
                         progress_loc=None,
                         problem_type='regression')

    # Test matches / works with cross_val_score
    cv_scores = cross_val_score(pipeline=pipe,
                                dataset=dataset,
                                scorer='r2',
                                problem_spec='default',
                                problem_type='regression')
    assert np.sum(evaluator.scores['r2']) == np.sum(cv_scores)

    assert len(evaluator.scores) > 0
    assert len(evaluator.timing['fit']) > 0
    assert len(evaluator.timing['score']) > 0
    evaluator._estimators_check()

    fis_df = evaluator.get_fis()
    assert list(fis_df) == ['1', '2']
    assert len(fis_df) == 5
    assert len(evaluator.feature_importances_) == 2

    raw_fis = evaluator.get_feature_importances()
    assert len(raw_fis) == 5
    assert evaluator.coef_ is None

    assert isinstance(evaluator.preds, dict)
    assert len(evaluator.preds['predict']) == 5
    assert len(evaluator.preds['y_true']) == 5

    assert len(evaluator.train_indices) == 5
    assert len(evaluator.val_indices) == 5


def test_evaluate_with_resid_grid_search():

    dataset = get_fake_dataset()
    resid = LinearResidualizer(to_resid_df=dataset[['1']])

    model = Model('dt', params={'criterion': Choice(['mse', 'friedman_mse'])})
    pipe = Pipeline(steps=[Scaler(obj=resid, scope='all'),
                           model], param_search=ParamSearch('grid'))

    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         problem_spec='default',
                         cv=5,
                         scorer='r2',
                         progress_bar=False,
                         store_preds=True,
                         store_estimators=True,
                         store_timing=True,
                         progress_loc=None,
                         problem_type='regression')

    assert isinstance(evaluator.estimators[0], BPtGridSearchCV)

    # Test matches / works with cross_val_score
    cv_scores = cross_val_score(pipeline=pipe,
                                dataset=dataset,
                                scorer='r2',
                                problem_spec='default',
                                problem_type='regression')
    assert np.sum(evaluator.scores['r2']) == np.sum(cv_scores)

    assert len(evaluator.scores) > 0
    assert len(evaluator.timing['fit']) > 0
    assert len(evaluator.timing['score']) > 0
    evaluator._estimators_check()

    fis_df = evaluator.get_fis()
    assert list(fis_df) == ['1', '2']
    assert len(fis_df) == 5
    assert len(evaluator.feature_importances_) == 2

    raw_fis = evaluator.get_feature_importances()
    assert len(raw_fis) == 5
    assert evaluator.coef_ is None

    assert isinstance(evaluator.preds, dict)
    assert len(evaluator.preds['predict']) == 5
    assert len(evaluator.preds['y_true']) == 5

    assert len(evaluator.train_indices) == 5
    assert len(evaluator.val_indices) == 5


def test_evaluate_compare():

    dataset = get_fake_binary_dataset()

    pipe1 = ModelPipeline(model=Model('dt'))
    pipe2 = ModelPipeline(model=Model('linear'))

    pipe = Compare([Option(pipe1, 'pipe1'),
                    Option(pipe2, 'pipe2')])

    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         progress_bar=False,
                         problem_type='binary')

    assert isinstance(evaluator, CompareDict)

    e1 = evaluator['pipe1']
    assert isinstance(e1, BPtEvaluator)
    e1_model = e1.estimators[0].steps[-1][1].estimator
    assert isinstance(e1_model, DecisionTreeClassifier)


def test_evaluate_compare_ps():

    dataset = get_fake_binary_dataset()

    pipe = ModelPipeline(model=Model('dt'))

    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         scope=Compare(['1', '2']),
                         progress_bar=False)

    assert isinstance(evaluator, CompareDict)


def test_evaluate_extra_params_compare():

    dataset = get_fake_dataset()

    pipe = ModelPipeline(model=Model('dt', max_depth=Compare([1, 2, 3])))
    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         progress_bar=False,
                         problem_type='regression')

    assert isinstance(evaluator, CompareDict)

    # Try different index methods
    e1 = evaluator['1']
    e2 = evaluator['model__max_depth=2']
    e3 = evaluator[Options(model__max_depth=3)]
    e4 = evaluator[Options(max_depth=3)]
    e5 = evaluator['max_depth=2']

    # Make sure actually set extra param
    es = [e1, e2, e3, e4, e5]
    for md, e in zip([1, 2, 3, 3, 2], es):
        assert e.estimators[0].steps[-1][1].estimator.max_depth == md


def test_evaluate_step_compare():

    dataset = get_fake_dataset()
    pipe = Pipeline(steps=[Compare([Model('dt'), Model('linear')])])

    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         progress_bar=False,
                         problem_type='regression')
    assert isinstance(evaluator, CompareDict)


def test_evaluate_step_compare_mp():

    dataset = get_fake_dataset()
    pipe = ModelPipeline(model=Compare([Model('dt'), Model('linear')]))

    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         progress_bar=False,
                         problem_type='regression')
    assert isinstance(evaluator, CompareDict)


def test_evaluate_step_compare2():

    dataset = get_fake_dataset()
    pipe = Pipeline(steps=[Compare([Scaler('standard'), Scaler('robust')]),
                           Model('dt')])

    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         progress_bar=False,
                         problem_type='regression')
    assert isinstance(evaluator, CompareDict)


def test_evaluate_step_compare2_mp():

    dataset = get_fake_dataset()
    pipe = ModelPipeline(scalers=Compare([Scaler('standard'),
                                          Scaler('robust')]),
                         model=Model('dt'))

    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         progress_bar=False,
                         problem_type='regression')
    assert isinstance(evaluator, CompareDict)


def test_evaluate_step_compare2_mp2():

    dataset = get_fake_dataset()
    pipe = ModelPipeline(scalers=[Scaler('standard'),
                                  Compare([Scaler('standard'),
                                           Scaler('robust')])],
                         model=Model('dt'))

    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         progress_bar=False,
                         problem_type='regression')
    assert isinstance(evaluator, CompareDict)


def test_evaluator_get_X_transform_df():

    pipe = Pipeline([Scaler('standard'), Model('dt')])
    dataset = get_fake_dataset()
    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         progress_bar=False,
                         cv=3)

    for fold in [0, 1, 2]:
        X_tr_trans = evaluator.get_X_transform_df(
            dataset, fold=fold, subjects='tr')
        assert X_tr_trans.sum().sum() == 0
        assert '1' in X_tr_trans
        assert '2' in X_tr_trans

        X_val_trans = evaluator.get_X_transform_df(
            dataset, fold=fold, subjects='val')
        assert X_val_trans.sum().sum() == 0
        assert '1' in X_val_trans
        assert '2' in X_val_trans

        assert len(X_tr_trans) + len(X_val_trans) == len(dataset)


def test_evaluate_pipeline_with_select():

    select_scaler = Select([Scaler('standard'), Scaler('robust')])
    select_model = Select([Model('linear'), Model('random forest')])

    pipe = Pipeline([select_scaler, select_model],
                    param_search=ParamSearch(n_iter=2))

    dataset = get_fake_dataset()
    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         progress_bar=False,
                         cv=3)

    assert isinstance(evaluator, BPtEvaluator)
    search_est = evaluator.estimators[0]
    assert isinstance(search_est, NevergradSearchCV)
    best_est = search_est.best_estimator_
    assert isinstance(best_est, BPtPipeline)
    step0 = best_est.steps[0]
    step1 = best_est.steps[1]
    assert isinstance(step0[1], Selector)
    assert isinstance(step1[1], Selector)
    assert step0[1].estimator_ == step0[1].estimators[step0[1].to_use][1]
    assert step1[1].estimator_ == step1[1].estimators[step1[1].to_use][1]


def test_evaluate_modelpipeline_with_select():

    select_scaler = Select([Scaler('standard'), Scaler('robust')])
    select_model = Select([Model('linear'), Model('random forest')])

    pipe = ModelPipeline(scalers=select_scaler,
                         model=select_model,
                         param_search=ParamSearch(n_iter=2))

    dataset = get_fake_dataset()
    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         progress_bar=False,
                         cv=3)

    assert isinstance(evaluator, BPtEvaluator)
    search_est = evaluator.estimators[0]
    assert isinstance(search_est, NevergradSearchCV)
    best_est = search_est.best_estimator_
    assert isinstance(best_est, BPtPipeline)
    step0 = best_est.steps[0]
    step1 = best_est.steps[1]
    assert isinstance(step0[1], Selector)
    assert isinstance(step1[1], Selector)
    assert step0[1].estimator_ == step0[1].estimators[step0[1].to_use][1]
    assert step1[1].estimator_ == step1[1].estimators[step1[1].to_use][1]


def test_evaluate_pipeline_with_custom_selector():

    feat_selector = FeatSelector('selector', params=1)
    model = Model('linear')

    pipe = Pipeline(steps=[feat_selector, model],
                    param_search=ParamSearch(n_iter=2))

    dataset = get_fake_dataset()
    evaluator = evaluate(pipeline=pipe,
                         dataset=dataset,
                         progress_bar=False,
                         cv=3)

    est1 = evaluator.estimators[0]
    assert isinstance(est1, NevergradSearchCV)

    b_est1 = est1.best_estimator_
    assert isinstance(b_est1, BPtPipeline)

    selector = b_est1[0]
    custom_sel = selector.estimator_

    from ...extensions.FeatSelectors import FeatureSelector as FS
    assert isinstance(custom_sel, FS)
    assert len(custom_sel.mask) == 2
    assert np.array_equal(custom_sel.mask, selector._get_support_mask())


def run_fs_checks(evaluator):

    est1 = evaluator.estimators[0]
    assert isinstance(est1, NevergradSearchCV)

    b_est1 = est1.best_estimator_
    assert isinstance(b_est1, BPtPipeline)

    ohe_tr = b_est1[0]
    assert ohe_tr.mapping_ == {0: 0, 1: 1}
    assert ohe_tr.out_mapping_ == {0: [0, 1], 1: 2}

    sel = b_est1[1]
    assert sel.mapping_ == {0: [0, 1], 1: 2}

    custom_sel = sel.estimator_
    assert custom_sel.mapping_ == {0: [0, 1], 1: 2}

    # Should be same as map to same spot
    msk = custom_sel.mask
    assert msk[0] == msk[1]

    if msk[0]:
        assert sel.out_mapping_[0] == 0
        assert sel.out_mapping_[1] == 1

        if msk[2]:
            assert sel.out_mapping_[2] == 2
            assert len(evaluator.feat_names[0]) == 3
        else:
            assert sel.out_mapping_[2] is None
            assert len(evaluator.feat_names[0]) == 2

    else:
        assert sel.out_mapping_[0] is None
        assert sel.out_mapping_[1] is None
        assert sel.out_mapping_[2] == 0
        assert len(evaluator.feat_names[0]) == 1

        if not msk[2]:
            raise AssertionError('Mask all False!')


def test_evaluate_pipeline_with_custom_selector_mapping():

    fake = Dataset()

    fake['1'] = np.ones((20))
    fake['1'][:10] = 0
    fake['2'] = np.ones((20))
    fake['3'] = np.ones((20))
    fake = fake.set_role('3', 'target')

    ohe = Transformer('one hot encoder', scope='1')
    feat_selector = FeatSelector('selector', params=1)
    model = Model('linear')

    pipe = Pipeline(steps=[ohe, feat_selector, model],
                    param_search=ParamSearch(n_iter=1))

    for rs in range(10):
        evaluator = evaluate(pipeline=pipe,
                             dataset=fake,
                             random_state=rs,
                             progress_bar=False,
                             cv=3)
        run_fs_checks(evaluator)





from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np
import warnings
import networkx as nx
from scipy.linalg import lstsq


class Identity(BaseEstimator, TransformerMixin):
    '''This loader will simply flatten the input object, if not already.
    This loader is used to for example pass along loaded surfaces
    or volumes to a PCA or other similar transformation.

    This object is designed to be used with input class :class:`BPt.Loader`
    for operating on single subjects at a time.
    '''

    def __init__(self):
        pass

    def fit(self, X, y=None):
        '''Fit accepts anything for X,
        and doesn't do anything except save the
        original shape of X.

        Parameters
        ----------
        X : numpy.array
            numpy.array with any shape, for one subject.

        y : numpy.array, optional
            This parameter is skipped.

            ::

                default = None

        '''
        self.X_shape_ = X.shape
        return self

    def fit_transform(self, X, y=None):
        '''Calls fit then transform, and returns the transformed output.

        Parameters
        ----------
        X : numpy.array
            numpy.array with any shape, for one subject.

        y : numpy.array, optional
            This parameter is skipped, it exists
            for compatibility.

            ::

                default = None

        '''
        return self.fit(X).transform(X)

    def transform(self, X):
        '''Transform simply returns a flattened version of the passed
        X, making it compatible with downstream classifiers.

        Parameters
        ----------
        X : numpy.array
            numpy.array with any shape, for one subject.

        Returns
        ---------
        X_trans : numpy.array
            1D flattened array for this subject.

        '''
        return X.flatten()

    def inverse_transform(self, X_trans):
        '''Inverse transform, i.e., un-flatten.

        Parameters
        ----------
        X_trans : numpy.array
            1D transformed numpy.array.

        Returns
        ---------
        X : numpy.array
            Data in original shape.

        '''

        return X_trans.reshape(self.X_shape_)


def load_surf(surf):
    '''Helper function to load a surface within BPt, w/ appropriate
    checks for important'''

    # If str, assume file path
    if isinstance(surf, str):

        # Only try to load with numpy if str ends with numpy ext
        if surf.endswith('.npy') or surf.endswith('.npz'):
            surf = np.load(surf)
            return surf

        # If not numpy, or numpy failed, try nilearn load surf data
        try:
            from nilearn.surface import load_surf_data
        except ImportError:
            raise ImportError('nilearn does not appear to be installed! ' +
                              'Install with "pip install nilearn", to load ' +
                              'surfaces from a file path.')

        surf = load_surf_data(surf)
        return surf

    # Keep as None for None
    elif surf is None:
        return None

    # If a Parc object
    elif hasattr(surf, 'get_parc'):
        return surf.get_parc(copy=True)

    # Otherwise assume either valid array-like passed
    else:
        return np.array(surf).copy()


class SurfLabels(BaseEstimator, TransformerMixin):
    '''Extract signals from non-overlapping labels.

    This class functions similar to
    :class:`nilearn.input_data.NiftiLabelsMasker`, except
    it is designed to work for surface, cifti or any arbitrary
    1 or 2D numpy arrays.

    Parameters
    -----------
    labels : str or array-like
        This should represent an array, of the same size as the data
        dimension, as a mask
        with unique integer values for each ROI. You can also pass a str
        location in which
        to load in this array (though the saved file must be loadable by
        either numpy.load, or
        if not a numpy array, will try and load with
        nilearn.surface.load_surf_data(), which you
        will need nilearn installed to use.)

    background_labels : int, array-like of int, optional
        This parameter determines which label, if any,
        in the corresponding
        passed labels, should be treated as 'background'
        and therefore no ROI
        calculated for that value or values.
        You may pass either a single interger
        value, an array-like of integer values.

        If not background label is desired, just pass
        a label which doesn't exist in any of the data,
        e.g., -100.

        ::

            default = 0

    mask : None, str or array-like, optional
        This parameter allows you to optional pass a mask of values in
        which to not calculate ROI values for.
        This can be passed as a str or
        array-like of values (just like labels),
        and should be comprised of
        a boolean array (or 1's and 0's),
        where a value of 1 means that value
        will be ignored (set to background label)
        should be kept, and a value of 0,
        for that value should be masked away.
        This array should have the same length as the passed `labels`.

        ::

            default = None

    strategy: specific str, custom_func, optional
        This parameter dictates the function to be applied
        to each data's ROI's
        individually, e.g., mean to calculate the mean by ROI.

        If a str is passed, it must correspond to one of the below preset
        options:

        - 'mean'
            Calculate the mean with :func:`numpy.mean`

        - 'sum'
            Calculate the sum with :func:`numpy.sum`

        - 'min' or 'minimum
            Calculate the min value with :func:`numpy.min`

        - 'max' or 'maximum
            Calculate the max value with :func:`numpy.max`

        - 'std' or 'standard_deviation'
            Calculate the standard deviation with :func:`numpy.std`

        - 'var' or  'variance'
            Calculate the variance with :func:`numpy.var`

        If a custom function is passed, it must accept two arguments,
        custom_func(X_i, axis=data_dim), X_i, where X_i is a subjects data
        array where that subjects data corresponds to
        labels == some class i, and can potentially
        be either a 1D array or 2D array, and an axis argument
        to specify which axis is
        the data dimension (e.g., if calculating for a time-series
        [n_timepoints, data_dim], then data_dim = 1,
        if calculating for say stacked contrasts where
        [data_dim, n_contrasts], data_dim = 0, and lastly for a 1D
        array, data_dim is also 0.

        ::

            default = 'mean'

    vectorize : bool, optional
        If the returned array should be flattened to 1D. E.g., if the
        last step in a set of loader steps this should be True, if before
        a different step it may make sense to set to False.

        ::

            default = True

    See Also
    -----------
    SurfMaps : For extracting non-static / probabilistic parcellations.
    nilearn.input_data.NiftiLabelsMasker : For working with volumetric data.

    '''

    def __init__(self, labels,
                 background_label=0,
                 mask=None,
                 strategy='mean',
                 vectorize=True):

        self.labels = labels
        self.background_label = background_label
        self.mask = mask
        self.strategy = strategy
        self.vectorize = vectorize

    def fit(self, X, y=None):
        '''Fit this object according
        the passed subjects data, X.

        Parameters
        ----------
        X : numpy.array
            numpy.array as either a 1D array,
            or a 2D array, where one dimension
            is matches the shape of the past
            labels and the other represents
            different time-points or modalities.

        y : numpy.array, optional
            This parameter is skipped.

            ::

                default = None
        '''

        # Load mask if any
        self.mask_ = load_surf(self.mask)

        # Load labels
        self.labels_ = load_surf(self.labels)

        if len(self.labels_.shape) > 1:
            raise RuntimeError('The passed labels array must be flat, ' +
                               ' i.e., a single dimension')

        # Mask labels if mask
        if self.mask_ is not None:

            # Raise error if wrong shapes
            if len(self.mask_) != len(self.labels_):
                raise RuntimeError('length of mask must have '
                                   'the same length / shape as '
                                   'the length of labels!')
            self.labels_[self.mask_.astype(bool)] = self.background_label

        # X can either be a 1D surface, or a 2D surface
        # (e.g. - for timeseries or stacked contrasts)
        if len(X.shape) > 2:
            raise RuntimeError('X can be at most 2D.')

        if len(self.labels_) not in X.shape:
            raise RuntimeError('Size of labels not found in X. '
                               'Make sure your data is in the same '
                               'space as the labels you are using!')

        # Proc self.background_label, if int turn to np array
        if isinstance(self.background_label, int):
            self.background_label_ = np.array([self.background_label])

        # Otherwise, if already list-like, just cast to np array
        else:
            self.background_label_ = np.array(self.background_label)

        # Set the _non_bkg_unique as the valid labels to get ROIs for
        self.non_bkg_unique_ = np.setdiff1d(np.unique(self.labels_),
                                            self.background_label_)

        # Proc strategy if need be
        strats = {'mean': np.mean,
                  'median': np.median,
                  'sum': np.sum,
                  'minimum': np.min,
                  'min': np.min,
                  'maximum': np.max,
                  'max': np.max,
                  'standard_deviation': np.std,
                  'std': np.std,
                  'variance': np.var,
                  'var': np.var}

        if self.strategy in strats:
            self.strategy_ = strats[self.strategy]
        else:
            self.strategy_ = self.strategy

        return self

    def fit_transform(self, X, y=None):
        '''Fit, then transform this object.

        Parameters
        ----------
        X : numpy.array
            numpy.array as either a 1D array,
            or a 2D array, where one dimension
            is matches the shape of the past
            labels and the other represents
            different time-points or modalities.

        y : numpy.array, optional
            This parameter is skipped.

            ::

                default = None

        Returns
        ---------
        X_trans : numpy.array
            The transformed data, either as 1D
            array is passed 1D data, or 2D if passed
            vectorize=False and originally 2D data.
        '''
        return self.fit(X, y).transform(X)

    def _check_fitted(self):
        if not hasattr(self, "labels_"):
            raise ValueError('It seems that SurfLabels has not been fitted. '
                             'You must call fit() before calling transform()')

    def transform(self, X):
        '''Transform this the passed data.

        If X has the both the same dimension's, raise warning.

        Parameters
        ----------
        X : numpy.array
            numpy.array as either a 1D array,
            or a 2D array, where one dimension
            is matches the shape of the past
            labels and the other represents
            different time-points or modalities.

        y : numpy.array, optional
            This parameter is skipped.

            ::

                default = None

        Returns
        ---------
        X_trans : numpy.array
            The transformed data, either as 1D
            array is passed 1D data, or 2D if passed
            vectorize=False and originally 2D data.
        '''

        self._check_fitted()

        if len(X.shape) == 2 and (X.shape[0] == X.shape[1]):
            warnings.warn('X was passed with the same length' +
                          ' in each dimension, ' +
                          'Assuming that axis=0 is the data dimension' +
                          ' w/ vertex values')

        if len(X.shape) > 2:
            raise RuntimeError('The shape of X can be at most 2 dimensions.')

        # The data dimension is just the dimension with
        # the same len as the label
        self.data_dim_ = X.shape.index(len(self.labels_))
        self.X_shape_ = X.shape

        # Get the ROI value for each label
        X_trans = []
        for i in self.non_bkg_unique_:

            if self.data_dim_ == 0:
                X_i = X[self.labels_ == i]
            else:
                X_i = X[:, self.labels_ == i]

            X_trans.append(self.strategy_(X_i, axis=self.data_dim_))

        if self.data_dim_ == 1:
            X_trans = np.stack(X_trans, axis=1)
        else:
            X_trans = np.array(X_trans)

        # Return based on vectorizes
        if not self.vectorize:
            return X_trans

        # Save original shape if vectorize called,
        # used for reverse transform
        self.original_shape_ = X_trans.shape
        return X_trans.flatten()

    def inverse_transform(self, X_trans):
        '''Reverse the original transformation.

        Parameters
        ----------
        X_trans : numpy.array
            Data with the same number of outputted
            features as data transformed
            by this object, e.g., the original
            transformed data or corresponding
            feature importances.

        Returns
        --------
        X : numpy.array
            The reverse transformed data
            passed back in its original space.

        '''

        # Reverse the vectorize
        if self.vectorize:
            X_trans = X_trans.reshape(self.original_shape_)

        X = np.zeros(self.X_shape_, dtype=X_trans.dtype, order='C')

        if self.data_dim_ == 1:
            X_trans = np.rollaxis(X_trans, -1)
            X = np.rollaxis(X, -1)

        for i, label in enumerate(self.non_bkg_unique_):
            X[self.labels_ == label] = X_trans[i]

        if self.data_dim_ == 1:
            X = np.rollaxis(X, -1)

        return X


class SurfMaps(BaseEstimator, TransformerMixin):
    '''Extract signals from overlapping labels.

    This class functions similar to
    :class:`nilearn.input_data.NiftiMapsMasker`, except
    it is designed to work for surface, cifti or any arbitrary
    1 or 2D numpy arrays.

    This object calculates the signal for each of the
    passed maps as extracted from the input during fit,
    and returns for each map a value.

    Parameters
    -----------
    maps : str or array-like, optional
        This parameter represents the maps in which
        to apply to each surface, where the shape of
        the passed maps should be (# of vertex, # of maps)
        or in other words, the size of the data array in the first
        dimension and the number of maps
        (i.e., the number of outputted ROIs from fit)
        as the second dimension.

        You may pass maps as either an array-like,
        or the str file location of a numpy or other
        valid surface file format array in which to load.

    strategy : {'auto', 'ls', 'average'}, optional
        The strategy in which the maps are used to extract
        signal. If 'ls' is selected, which stands for
        least squares, the least-squares solution will
        be used for each region.

        Alternatively, if 'average' is passed, then
        the weighted average value for each map
        will be computed.

        By default 'auto' will be selected,
        which will use 'average' if the passed
        maps contain only positive weights, and
        'ls' in the case that there are
        any negative values in the passed maps.

        Otherwise, you can set a specific strategy.
        In deciding which method to use,
        consider an example. Let's say the
        fit data X, and maps are

        ::

            data = np.array([1, 1, 5, 5])
            maps = np.array([[0, 0],
                             [0, 0],
                             [1, -1],
                             [1, -1]])

        In this case, the 'ls' method would
        yield region signals [2.5, -2.5], whereas
        the weighted 'average' method, would yield
        [5, 5], notably ignoring the negative weights.
        This highlights an important limitation to the
        weighted averaged method, as it does not
        handle negative values well.

        On the other hand, consider changing the maps
        weights to

        ::

            data = np.array([1, 1, 5, 5])
            maps = np.array([[0, 1],
                             [0, 2],
                             [1, 0],
                             [1, 0]])

            ls_sol = [5. , 0.6]
            average_sol = [5, 1]

        In this case, we can see that the weighted
        average gives a maybe more intuitive summary
        of the regions. In general, it depends on
        what signal you are trying to summarize, and
        how you are trying to summarize it.

    mask : None, str or array-like, optional
        This parameter allows you to optional pass a mask of values in
        which to not calculate ROI values for.
        This can be passed as a str or
        array-like of values (just like maps),
        and should be comprised of
        a boolean array (or 1's and 0's),
        where a value of 1 means that value
        will be ignored (set to 0)
        should be kept, and a value of 0,
        for that value should be masked away.
        This array should have the same length as the passed `maps`.
        Specifically, where the shape of maps is (size, n_maps),
        the shape of mask should be (size).

        ::

            default = None

    vectorize : bool, optional
        If the returned array should be flattened to 1D. E.g., if this is
        the last step in a set of loader steps this should be True.
        Also note, if the surface data it is being applied to is 1D,
        then the output will be 1D regardless of this parameter.

        ::

            default = True

    See Also
    -----------
    SurfLabels : For extracting static / non-probabilistic parcellations.
    nilearn.input_data.NiftiMapsMasker : For volumetric nifti data.

    Examples
    ----------
    First let's define an example set of probabilistic maps, we
    will assume there are just 5 features in our data, and we will
    define 6 total maps.

    .. ipython:: python

        import numpy as np
        from BPt.extensions import SurfMaps

        # This should have shape number of features x number of maps!
        prob_maps = np.array([[3, 1, 1, 1, 1, 1],
                              [1, 3, 1, 1, 1, 1],
                              [1, 1, 3, 1, 1, 1],
                              [1, 1, 1, 3, 1, 1],
                              [1, 1, 1, 1, 3, 1]])
        prob_maps.shape

    Next we can define some input data to use with these maps.

    .. ipython:: python

        data1 = np.arange(5, dtype='float')
        data1
        data2 = np.ones(5, dtype='float')
        data2

    Now let's define the actual object and use it to transform the data.

    .. ipython:: python

        sm = SurfMaps(maps=prob_maps)
        sm.fit_transform(data1)
        sm.fit_transform(data2)

    Okay so what is going on when we transform this data? Basically we are
    just taking weighted averages for each one of the defined maps. We could
    also explicitly change the strategy from 'auto' to 'ls' which would
    take the least squares solution instead.

    .. ipython:: python

        sm = SurfMaps(maps=prob_maps, strategy='ls')
        data_trans = sm.fit_transform(data1)
        data_trans

    While a little less intuitive, the least squares solution allows
    us to reverse the feature transformation (although not always exactly)

    .. ipython:: python

        sm.inverse_transform(data_trans)

    This can be useful in the say the case of converting back downstream
    calculated feature importance to the original data space.
    '''

    def __init__(self, maps, strategy='auto', mask=None, vectorize=True):
        self.maps = maps
        self.strategy = strategy
        self.mask = mask
        self.vectorize = vectorize

    def fit(self, X, y=None):
        '''Fit this object according
        the passed subjects data, X.

        Parameters
        ----------
        X : numpy.array
            numpy.array as either a 1D array,
            or a 2D array, where one dimension
            is matches the shape of the past
            labels and the other represents
            different time-points or modalities.

        y : numpy.array, optional
            This parameter is skipped.

            ::

                default = None
        '''

        # Load mask if any
        self.mask_ = load_surf(self.mask)

        # Load maps
        self.maps_ = load_surf(self.maps)

        # Save dtype
        self.dtype_ = X.dtype

        # Warn if non-float
        if 'float' not in self.dtype_.name:
            warnings.warn('The original datatype is non-float, ' +
                          'this may lead to rounding errors! ' +
                          'Pass data as type float to ensure ' +
                          'the results of transform are not truncated.')

        # Make the maps if passed mask set to 0 in those spots
        if self.mask_ is not None:

            # Raise error if wrong shapes
            if len(self.mask_) != self.maps_.shape[0]:
                raise RuntimeError('length of mask must have '
                                   'the same length / shape as '
                                   'the first dimension of passed '
                                   'maps!')
            self.maps_[self.mask_.astype(bool)] = 0

        # X can either be a 1D surface, or a 2D surface
        # (e.g. - for timeseries or stacked contrasts)
        if len(X.shape) > 2:
            raise RuntimeError('X can be at most 2D.')

        # Check to make sure dimension of data is correct
        if self.maps_.shape[0] not in X.shape:
            raise RuntimeError('Size of labels not found in X. '
                               'Make sure your data is in the same '
                               'space as the labels you are using!')

        # Make sure strategy exists
        if self.strategy not in ['auto', 'ls', 'average']:
            raise RuntimeError('strategy must be '
                               '"ls", "average" or "auto"!')

        return self

    def fit_transform(self, X, y=None):
        '''Fit, then transform this object.

        Parameters
        ----------
        X : numpy.array
            numpy.array as either a 1D array,
            or a 2D array, where one dimension
            is matches the shape of the past
            labels and the other represents
            different time-points or modalities.

        y : numpy.array, optional
            This parameter is skipped.

            ::

                default = None

        Returns
        ---------
        X_trans : numpy.array
            The transformed data, either as 1D
            array is passed 1D data, or 2D if passed
            vectorize=False and originally 2D data.
        '''

        return self.fit(X, y).transform(X)

    def _check_fitted(self):
        if not hasattr(self, "maps_"):
            raise ValueError('It seems that SurfMaps has not been fitted. '
                             'You must call fit() before calling transform()')

    def transform(self, X):
        '''Transform this the passed data.

        If X has the both the same dimension's, raise warning.

        Parameters
        ----------
        X : numpy.array
            numpy.array as either a 1D array,
            or a 2D array, where one dimension
            is matches the shape of the past
            labels and the other represents
            different time-points or modalities.

        y : numpy.array, optional
            This parameter is skipped.

            ::

                default = None

        Returns
        ---------
        X_trans : numpy.array
            The transformed data, either as 1D
            array is passed 1D data, or 2D if passed
            vectorize=False and originally 2D data.
        '''

        self._check_fitted()

        if len(X.shape) == 2 and (X.shape[0] == X.shape[1]):
            warnings.warn('X was passed with the same length',
                          ' in each dimension, ',
                          'Assuming that axis=0 is the data dimension',
                          ' w/ vertex values')

        # The data dimension is just the dimension with
        # the first dimension of maps
        self.data_dim_ = X.shape.index(self.maps_.shape[0])
        self.X_shape_ = X.shape

        # If data in second dimension, transpose
        if self.data_dim_ == 1:
            X = X.T

        # Set strategy if auto
        self.strategy_ = self.strategy
        if self.strategy_ == 'auto':
            self.strategy_ = 'ls'
            if np.all(self.maps_ >= 0):
                self.strategy_ = 'average'

        # Run the correct transform based on strategy
        if self.strategy_ == 'ls':
            X_trans = self._transform_ls(X)
        elif self.strategy_ == 'average':
            X_trans = self._transform_average(X)
        else:
            X_trans = None

        # Convert back to original dtype
        X_trans = X_trans.astype(self.dtype_)

        # Always return as shape of extra data if any by
        # number of maps, with number of maps as last dimension

        # Return based on vectorize
        if not self.vectorize:
            return X_trans

        # Save original transformed output shape if vectorize
        self.original_shape_ = X_trans.shape
        return X_trans.flatten()

    def _transform_ls(self, X):
        '''X should be data points, or data points x stacked.'''

        X_trans = lstsq(self.maps_, X)[0]
        return X_trans.T

    def _transform_average(self, X):
        '''X should be data points, or data points x stacked.'''

        X_trans = []

        # For each map - take weighted average
        for m in range(self.maps_.shape[1]):

            try:
                X_trans.append(np.average(X, axis=0, weights=self.maps_[:, m]))
            except ZeroDivisionError:
                pass

        return np.array(X_trans).T

    def inverse_transform(self, X_trans):
        '''Reverse the original transformation.

        Parameters
        ----------
        X_trans : numpy.array
            Data with the same number of outputted
            features as data transformed
            by this object, e.g., the original
            transformed data or corresponding
            feature importances.

        Returns
        --------
        X : numpy.array
            The reverse transformed data
            passed back in its original space.

        '''

        # Reverse the vectorize, if needed
        if self.vectorize:
            X_trans = X_trans.reshape(self.original_shape_)

        if self.strategy_ == 'ls':
            return np.dot(X_trans, self.maps_.T)

        elif self.strategy_ == 'average':
            raise RuntimeError('Cannot calculate reverse of average.')


def proc_X(X):

    if not isinstance(X, list):
        if len(np.shape(X)) == 2:
            return [X]

    return X


def proc_X_trans(X_trans, vectorize):

    if X_trans.shape[0] == 1:
        X_trans = X_trans.reshape(X_trans.shape[1:])

    if vectorize:
        X_trans = X_trans.flatten()

    return X_trans


# Create wrapper for nilearn connectivity measure to make it
# work with 1 subject
try:
    from nilearn.connectome import ConnectivityMeasure

    class SingleConnectivityMeasure(ConnectivityMeasure):
        '''| See :class:`nilearn.connectome.ConnectivityMeasure`.
          This class is just a wrapper to let this object work when passed
          a single connectivity matrix.

        | This class requires extra dependency nilearn to be installed.
        '''

        def fit(self, X, y=None):
            return super().fit(proc_X(X), y)

        def fit_transform(self, X, y=None):
            X_trans = super().fit_transform(proc_X(X), y)
            return proc_X_trans(X_trans, self.vectorize)

        def transform(self, X):
            X_trans = super().transform(proc_X(X))
            return proc_X_trans(X_trans, self.vectorize)

except ImportError:
    pass


def avg(func):

    def mean_func(G):
        return np.mean(func(G).values())

    return mean_func


class ThresholdNetworkMeasures(BaseEstimator, TransformerMixin):
    '''This class is designed for thresholding and then extracting network
    measures from an input correlation matrix.

    Parameters
    -----------
    threshold : float, optional
        A floating point threshold between 0 and 1.
        This represents the threshold at which a connection
        in the passed data needs to be in order for it to
        be set as an edge. The type of threshold_method
        also changes how this threshold behaves.

        ::

            default = .2

    threshold_type : {'abs', 'pos', 'neg'}
        The type of thresholding, e.g., should the threshold
        be applied to:

        - 'abs'
            The absolute value of the connections

        - 'pos'
            Only consider edges as passed the threshold if >= self.threshold

        - 'neg'
            Only consider edges as passed the if <= self.threshold

        ::

            default = 'abs'

    threshold_method : {'value', 'density'}, optional
        The method for thresholding. The two
        defined options are either to define an edge
        strictly by value, e.g., if threshold_type is 'abs',
        and threshold is .2, then any connection greater than or
        equal to .2 will be set as an edge.

        Alternatively, you may specify that the threshold be
        treated as a density. What this means is that if the threshold
        is set to for example .2, then the top 20% of edges by weight
        will be set as an edge, regardless of the actual value of the edge.

        Note: The passed percentage will be considered
        out of all of the the total possible edges in the adjacency matrix.
        That includes both the upper and lower repeated triangles
        if non-directional edges.

        ::

            default = 'value'

    to_compute : valid_measure or list of, optional
        | Either a single str representing a network
            measure to compute, or a list of valid
            measures. You may also pass any custom function
            which accepts one argument G, and returns
            a value.

        | The following global measures are currently implemented
            as options:
        |


        - 'avg_cluster':
            :func:`networkx.algorithms.cluster.average_clustering`
        - 'assortativity':
            :func:`networkx.algorithms.assortativity.degree_assortativity_coefficient`
        - 'global_eff':
            :func:`networkx.algorithms.efficiency_measures.global_efficiency`
        - 'local_eff':
            :func:`networkx.algorithms.efficiency_measures.local_efficiency`
        - 'sigma':
            :func:`networkx.algorithms.smallworld.sigma`
        - 'omega`:
            :func:`networkx.algorithms.smallworld.omega`
        - 'transitivity':
            :func:`networkx.algorithms.cluster.transitivity`

        |
        | You may also select from one of the following
            averages of local measures:
        |

        - 'avg_eigenvector_centrality':
            :func:`networkx.algorithms.centrality.eigenvector_centrality_numpy`
        - 'avg_closeness_centrality':
            :func:`networkx.algorithms.centrality.closeness_centrality`
        - 'avg_degree':
            Average graph degree.
        - 'avg_triangles':
            :func:`networkx.algorithms.cluster.triangles`
        - 'avg_pagerank':
            :func:`networkx.algorithms.link_analysis.pagerank_alg.pagerank`
        - 'avg_betweenness_centrality':
            :func:`networkx.algorithms.centrality.betweenness_centrality`
        - 'avg_information_centrality':
            :func:`networkx.algorithms.centrality.information_centrality`
        - 'avg_shortest_path_length':
            :func:`networkx.algorithms.shortest_paths.generic.average_shortest_path_length`


        ::

            default = 'avg_degree'

    '''

    # @TODO DIGRAPH CASE??

    def __init__(self, threshold=.2,
                 threshold_type='abs',
                 threshold_method='value',
                 to_compute='avg_degree'):

        self.threshold = threshold
        self.threshold_type = threshold_type
        self.threshold_method = threshold_method
        self.to_compute = to_compute

    @property
    def feat_names_(self):
        '''The list of feature names returned
        by this objects transform function. This property
        is special in that it can interact with :class:`BPt.Loader`,
        passing along feature name information.
        '''
        return self._feat_names

    @feat_names_.setter
    def feat_names_(self, feat_names):
        self._feat_names = feat_names

    def fit(self, X, y=None):
        '''X is a 2d correlation matrix'''

        if isinstance(self.to_compute, str):
            self.to_compute = [self.to_compute]

        try:
            import networkx
            networkx
        except ImportError:
            raise ImportError(
                'To use this class, make sure you have networkx installed!')

        # The dictionary of valid options
        self._func_dict = {
            'avg_cluster': nx.average_clustering,
            'assortativity': nx.degree_assortativity_coefficient,
            'global_eff': nx.global_efficiency,
            'local_eff': nx.local_efficiency,
            'sigma': nx.sigma,
            'omega': nx.omega,
            'transitivity': nx.transitivity,
            'avg_eigenvector_centrality': avg(nx.eigenvector_centrality_numpy),
            'avg_closeness_centrality': avg(nx.closeness_centrality),
            'avg_degree': self._avg_degree,
            'avg_triangles': avg(nx.triangles),
            'avg_pagerank': avg(nx.pagerank),
            'avg_betweenness_centrality': avg(nx.betweenness_centrality),
            'avg_information_centrality': avg(nx.information_centrality),
            'avg_shortest_path_length': nx.average_shortest_path_length
        }

        # Compute the feat names to return once here.
        self.feat_names_ = []
        for compute in self.to_compute:
            if compute not in self._func_dict:
                self.feat_names_.append(compute.__name__)
            else:
                self.feat_names_.append(compute)

        return self

    def fit_transform(self, X, y=None):
        '''Fit, then transform a passed 2D numpy correlation matrix.

        Parameters
        ----------
        X : numpy array
            A 2D numpy array representing an input correlation
            matrix.

        Returns
        ---------
        X_trans : numpy array
            Returns a flat array of length number of
            measures in parameter to_compute, representing
            the calculated network statistics.
        '''

        return self.fit(X, y).transform(X)

    def _apply_threshold(self, X):

        # Process threshold type on copy of X
        X_t = X.copy()
        if self.threshold_type == 'abs':
            X_t = np.abs(X_t)

        # If Value
        if self.threshold_method == 'value':
            if self.threshold_type == 'neg':
                return np.where(X_t <= self.threshold, 1, 0)
            return np.where(X_t >= self.threshold, 1, 0)

        elif self.threshold_method == 'density':

            top_n = round(len(X_t) * self.threshold) - 1

            # If less than 0, set to 0
            if top_n < 0:
                top_n = 0

            reverse = False if self.threshold_type == 'neg' else True
            thresh = sorted(np.triu(X_t).flatten(), reverse=reverse)[top_n]
            return np.where(X_t >= thresh, 1, 0)

        raise RuntimeError(str(self.threshold_method) + ' not a valid.')

    def _threshold_check(self, X):

        while np.sum(self._apply_threshold(X)) == 0:
            warnings.warn('Setting threshold lower than: ' +
                          str(self.threshold) + '. As, otherwise no edges ' +
                          'will be set. New threshold = ' +
                          str(self.threshold - .01))
            self.threshold -= .01

    def transform(self, X):
        '''Transform a passed 2D numpy correlation matrix.

        Parameters
        ----------
        X : numpy array
            A 2D numpy array representing an input correlation
            matrix.

        Returns
        ---------
        X_trans : numpy array
            Returns a flat array of length number of
            measures in parameter to_compute, representing
            the calculated network statistics.
        '''

        # Make sure the specified threshold doesn't break everything
        self._threshold_check(X)

        # Apply threshold
        X = self._apply_threshold(X)
        G = nx.from_numpy_matrix(X)

        X_trans = []

        for compute in self.to_compute:
            if compute not in self._func_dict:
                X_trans += [compute(G)]
            else:
                X_trans += [self._func_dict[compute](G)]
        return np.array(X_trans)

    def _avg_degree(self, G):
        avg_degree = np.mean([i[1] for i in nx.degree(G)])
        return avg_degree



from sklearn.experimental import enable_hist_gradient_boosting
from ...extensions.MLP import MLPRegressor_Wrapper, MLPClassifier_Wrapper
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.naive_bayes import GaussianNB
from sklearn.gaussian_process import (GaussianProcessClassifier,
                                      GaussianProcessRegressor)
from sklearn.ensemble import (GradientBoostingClassifier,
                              GradientBoostingRegressor,
                              RandomForestRegressor, RandomForestClassifier,
                              ExtraTreesClassifier, ExtraTreesRegressor,
                              HistGradientBoostingRegressor,
                              HistGradientBoostingClassifier)
from sklearn.linear_model import (LogisticRegression, ElasticNet,
                                  LinearRegression, HuberRegressor,
                                  Lasso, Ridge, SGDClassifier,
                                  SGDRegressor,
                                  PassiveAggressiveClassifier,
                                  BayesianRidge, ARDRegression,
                                  TweedieRegressor)

from sklearn.svm import SVC, LinearSVR, SVR, LinearSVC
from ..helpers import get_obj_and_params, all_from_avaliable


AVALIABLE = {
    'binary': {
        'logistic':           'logistic',
        'linear':             'logistic',
        'lasso':              'lasso logistic',
        'ridge':              'ridge logistic',
        'elastic':            'elastic net logistic',
        'elastic net':        'elastic net logistic',
        'gaussian nb':        'gaussian nb',
        'knn':                'knn classifier',
        'dt':                 'dt classifier',
        'rf':                 'random forest classifier',
        'random forest':      'random forest classifier',
        'gp':                 'gp classifier',
        'svm':                'svm classifier',
        'svc':                'svm classifier',
        'linear svm':         'linear svm classifier',
        'linear svc':         'linear svm classifier',
        'mlp':                'mlp classifier',
        'sgd':                'sgd classifier',
        'gb':                 'gb classifier',
        'hgb':                'hgb classifier',
        'et':                 'et classifier',
        'pa':                 'pa classifier',

    },
    'regression': {
        'linear':             'linear regressor',
        'knn':                'knn regressor',
        'dt':                 'dt regressor',
        'elastic':            'elastic net regressor',
        'elastic net':        'elastic net regressor',
        'rf':                 'random forest regressor',
        'random forest':      'random forest regressor',
        'gp':                 'gp regressor',
        'svm':                'svm regressor',
        'svr':                'svm regressor',
        'linear svm':         'linear svm regressor',
        'linear svr':         'linear svm regressor',
        'mlp':                'mlp regressor',
        'ridge':              'ridge regressor',
        'lasso':              'lasso regressor',
        'gb':                 'gb regressor',
        'hgb':                'hgb regressor',
        'et':                 'et regressor',
        'bayesian ridge':     'bayesian ridge regressor',
        'ard':                'ard regressor',
        'tweedie':            'tweedie regressor',
    },
}

# Should be the same
AVALIABLE['categorical'] = AVALIABLE['binary'].copy()


MODELS = {
    'logistic': (LogisticRegression, ['base logistic']),

    'lasso logistic': (LogisticRegression, ['base lasso', 'lasso C',
                                            'lasso C extra']),

    'ridge logistic': (LogisticRegression, ['base ridge', 'ridge C',
                                            'ridge C extra']),

    'elastic net logistic': (LogisticRegression, ['base elastic',
                                                  'elastic classifier',
                                                  'elastic clf v2',
                                                  'elastic classifier extra']),

    'elastic net regressor': (ElasticNet, ['base elastic net',
                                           'elastic regression',
                                           'elastic regression extra']),

    'ridge regressor': (Ridge, ['base ridge regressor',
                                'ridge regressor dist']),
    'lasso regressor': (Lasso, ['base lasso regressor',
                                'lasso regressor dist']),

    'huber': (HuberRegressor, ['base huber']),

    'gaussian nb': (GaussianNB, ['base gnb']),

    'knn classifier': (KNeighborsClassifier, ['base knn', 'knn dist']),
    'knn regressor': (KNeighborsRegressor, ['base knn regression',
                                            'knn dist regression']),

    'dt classifier': (DecisionTreeClassifier, ['default',
                                               'dt classifier dist']),
    'dt regressor':  (DecisionTreeRegressor, ['default', 'dt dist']),

    'linear regressor': (LinearRegression, ['base linear']),

    'random forest regressor': (RandomForestRegressor, ['base rf', 'rf dist']),
    'random forest classifier': (RandomForestClassifier,
                                 ['base rf regressor', 'rf classifier dist']),

    'gp regressor': (GaussianProcessRegressor, ['base gp regressor']),
    'gp classifier': (GaussianProcessClassifier,  ['base gp classifier']),

    'svm regressor': (SVR, ['base svm', 'svm dist']),
    'svm classifier': (SVC, ['base svm classifier', 'svm classifier dist']),

    'mlp regressor': (MLPRegressor_Wrapper, ['default', 'mlp dist 3 layer',
                                             'mlp dist es 3 layer',
                                             'mlp dist 2 layer',
                                             'mlp dist es 2 layer',
                                             'mlp dist 1 layer',
                                             'mlp dist es 1 layer']),
    'mlp classifier': (MLPClassifier_Wrapper, ['default', 'mlp dist 3 layer',
                                               'mlp dist es 3 layer',
                                               'mlp dist 2 layer',
                                               'mlp dist es 2 layer',
                                               'mlp dist 1 layer',
                                               'mlp dist es 1 layer']),

    'linear svm classifier': (LinearSVC, ['base linear svc',
                                          'linear svc dist']),
    'linear svm regressor': (LinearSVR, ['base linear svr',
                                         'linear svr dist']),

    'sgd classifier': (SGDClassifier, ['base sgd', 'sgd elastic classifier',
                                       'sgd classifier big search']),
    'sgd regressor': (SGDRegressor, ['base sgd', 'sgd elastic']),

    'gb classifier': (GradientBoostingClassifier, ['default']),
    'gb regressor': (GradientBoostingRegressor, ['default']),

    'hgb classifier': (HistGradientBoostingClassifier, ['default']),
    'hgb regressor': (HistGradientBoostingRegressor, ['default']),

    'et classifier': (ExtraTreesClassifier, ['default']),
    'et regressor': (ExtraTreesRegressor, ['default']),

    'pa classifier': (PassiveAggressiveClassifier, ['default']),

    'bayesian ridge regressor': (BayesianRidge, ['default']),

    'ard regressor': (ARDRegression, ['default']),

    'tweedie regressor': (TweedieRegressor, ['default']),
}

try:
    from xgboost import XGBClassifier, XGBRegressor

    AVALIABLE['binary']['xgb'] = 'xgb classifier'
    AVALIABLE['regression']['xgb'] = 'xgb regressor'
    AVALIABLE['categorical']['xgb'] = 'xgb classifier'

    MODELS['xgb regressor'] = (XGBRegressor, ['base xgb', 'xgb dist1',
                                              'xgb dist2', 'xgb dist3'])
    MODELS['xgb classifier'] = (XGBClassifier, ['base xgb classifier',
                                                'xgb classifier dist1',
                                                'xgb classifier dist2',
                                                'xgb classifier dist3'])

except ImportError:
    pass

try:
    from ...extensions.BPtLGBM import BPtLGBMRegressor, BPtLGBMClassifier

    AVALIABLE['binary']['light gbm'] = 'light gbm classifier'
    AVALIABLE['binary']['lgbm'] = 'light gbm classifier'
    AVALIABLE['categorical']['light gbm'] = 'light gbm classifier'
    AVALIABLE['categorical']['lgbm'] = 'light gbm classifier'
    AVALIABLE['regression']['light gbm'] = 'light gbm regressor'
    AVALIABLE['regression']['lgbm'] = 'light gbm regressor'

    MODELS['light gbm regressor'] = (BPtLGBMRegressor, ['base lgbm',
                                                        'lgbm dist1',
                                                        'lgbm dist2',
                                                        'lgbm dist3'])
    MODELS['light gbm classifier'] = (BPtLGBMClassifier,
                                      ['base lgbm',
                                       'lgbm classifier dist1',
                                       'lgbm classifier dist2',
                                       'lgbm classifier dist3'])
except ImportError:
    pass


def get_base_model_and_params(model_type, extra_params, model_type_params,
                              random_state=None, **kwargs):

    model, extra_model_params, model_type_params =\
        get_obj_and_params(model_type, MODELS, extra_params, model_type_params)

    return model(**extra_model_params), model_type_params


all_obj_keys = all_from_avaliable(AVALIABLE)
